#
# Copyright (c) Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import logging
from os2mo_helpers.mora_helpers import MoraHelper
from integration_abstraction.integration_abstraction import IntegrationAbstraction

from os2mo_data_import.utilities import ImportUtility
from os2mo_data_import.defaults import facet_defaults
from os2mo_data_import.mora_data_types import (
    mora_type_config,
    AddressType,
    EngagementType,
    AssociationType,
    RoleType,
    ManagerType,
    LeaveType,
    ItsystemType,
    OrganisationUnitType,
    EmployeeType,
    TerminationType,
    EngagementTerminationType
)

from os2mo_data_import.mox_data_types import (
    Organisation,
    Klassifikation,
    Facet,
    Klasse,
    Itsystem
)

logger = logging.getLogger("moImporterHelpers")


class ImportHelper:
    """
    The entry point of an import is the ImportHelper.

    Example:

    .. code-block:: python

        os2mo = ImportHelper(
            create_defaults=True,
            store_integration_data=True
        )

        os2mo.import_all()


    :param str system_name: An identifier for import and/or integration

    It will be stored as the main identifier for integration data,
    allowing for consecutive imports without causing conflicts in the database.

    :param str end_marker: A semi unique marker

    This marks the end of the system name value.

    .. note::
        Consider the following example,
        where the system name is "Import" and the value is "ACME TM".
        The endmarker defaults to "_|-STOP".

        The data (integration data) is stored as a json string:

        .. code-block:: text
            :emphasize-lines: 5

            {
                "organisationegenskaber": [
                    {
                      ...
                      "integrationsdata": "{\"Import\": \"ACME TM_|-STOP\"}",
                      ...
                  ]
            }

        This is an attempt to eliminate collisions with "common" values.

    :param str mox_base: The base url of the mox backend
    :param str mora_base: The base url of the mora backend

    :param bool store_integration_data: Store the integration data in the database

    Any integration should only import store_integration_data=True.

    :param bool create_defaults: Create the default set of "facet" types

    .. note::
        For more information on the actual types, please see the defaults module.

        :py:data:`os2mo_data_import.defaults`

    :param class ImportUtility: Default import class

    .. note::
        For more information see:

        :class:`os2mo_data_import.utilities.ImportUtility`
    """

    def __init__(self, system_name="Import", end_marker="_|-STOP",
                 mox_base="http://localhost:8080", mora_base="http://localhost:5000",
                 store_integration_data=False, create_defaults=True,
                 seperate_names=False, demand_consistent_uuids=True,
                 ImportUtility=ImportUtility):

        self.seperate_names = seperate_names
        mora_type_config(mox_base=mox_base,
                         system_name=system_name,
                         end_marker=end_marker)
        self.mox_base = mox_base
        # Import Utility
        self.store = ImportUtility(
            mox_base=mox_base,
            mora_base=mora_base,
            system_name=system_name,
            end_marker=end_marker,
            demand_consistent_uuids=demand_consistent_uuids,
            store_integration_data=store_integration_data
        )
        # TODO: store_integration_data could be passed to ImportUtility by passing
        # the actual self.ia object
        if store_integration_data:
            self.morah = MoraHelper(use_cache=False)
            self.ia = IntegrationAbstraction(mox_base, system_name, end_marker)

        self.organisation = None
        self.klassifikation = None

        self.klasse_objects = {}
        self.facet_objects = {}
        self.addresses = []
        self.itsystems = {}

        self.organisation_units = {}
        self.employees = {}

        # Compatibility map
        self.available_types = {
            "klasse": "klasse_objects",
            "facet": "facet_objects",
            "organisation_unit": "organisation_units",
            "employee": "employees"
        }

        self.organisation_unit_details = {}
        self.employee_details = {}

        # Create default facet and klasse
        if create_defaults:
            self.create_default_facet_types()

    def get(self, object_type, object_reference):
        """
        Get a mapped object by type and reference.

        Example:

        .. code-block:: python

            os2mo.get("employee", "Luke Skywalker")

        :param str object_type: Reference to type of the map

        This can be either of the following:

         - klasse
         - facet
         - organisation_unit
         - employee

        :param str object_reference: Reference to an identifier in the map

        :return: The selected object
        :rtype: object
        """
        available = self.export(object_type)
        return available.get(object_reference)

    def export(self, object_type):
        """
        Compatibitlity method to fetch all objects by type.

        :param str object_type: Reference to type of the map

        :return: A list containing all mapped objects with selected type
        :rtype: list<object>
        """
        available = self.available_types.keys()

        if object_type not in available:
            raise TypeError(
                "Cannot check for this type, available types: {}".format(
                    available
                )
            )

        object_type = self.available_types.get(object_type)

        return getattr(self, object_type)

    def check_if_exists(self, object_type, object_reference):
        """
        Helper method to check for the existence of a mapped object
        by type and reference.

        :param str object_type: Reference to type of the map
        :param str object_reference: Reference to an identifier in the map

        :return: Exists or does not exist
        :rtype: bool
        """

        attribute = self.export(object_type)

        if not attribute.get(object_reference):
            return False

        return True

    def get_details(self, owner_type, owner_ref, type_id=None):
        """
        Retrieve a list of details of either an employee or organisation unit.

        :param str owner_type: employee or organisation unit
        :param str owner_ref: Reference to the identifier of the owner
        :param str type_id: (optional) type of detail

        Example:

        .. code-block:: python

            os2mo.get_details("employee", "Luke Skywalker", "engagement")

        :return: A list of details
        :rtype: list<object>
        """

        if owner_type == "organisation_unit":
            details = self.organisation_unit_details[owner_ref]
        elif owner_type == "employee":
            details = self.employee_details[owner_ref]
        else:
            raise ReferenceError(
                "Owner type must be either employee or organisation_unit"
            )

        if not details:
            details = []

        if type_id:
            details = [
                detail
                for detail in details
                if detail.type_id == type_id
            ]

        return details

    def create_validity(self, date_from, date_to):
        """
        Shorthand to create validity from the date_from and date_to attributes.

        :param str date_from: Start date e.g. "1982-01-01"

        :param str date_to: End date e.g. "1982-01-01"

        :return: Validity
        :rtype: dict
        """

        if not date_from or not date_to:
            raise AssertionError("Date is not specified, cannot create validity")

        return {
            "from": date_from,
            "to": date_to
        }

    def add_organisation(self, identifier, **kwargs):
        """
        Add an Organisation object to the map

        :param str identifier: A user defined identifier
        :param kwargs kwargs: :class:`os2mo_data_import.mox_data_types.Organisation`

        .. note::
            If name is not set, the identifier is used as the name.
        """

        name = (
            kwargs.get("name") or identifier
        )

        self.organisation = (
            identifier,
            Organisation(name=name, **kwargs),
        )

        self.klassifikation = (
            identifier,
            Klassifikation(user_key=name, parent_name=name, description="umbrella")
        )

    def add_klasse(self, identifier, **kwargs):
        """
        Add a Klasse object to the map.

        :param str identifier: A user defined identifier
        :param kwargs kwargs: :class:`os2mo_data_import.mox_data_types.Klasse`
        """

        if identifier in self.klasse_objects:
            raise ReferenceError(
                'Unique constraint - Klasse identifier {} exists'.format(identifier)
            )

        if "user_key" not in kwargs:
            kwargs["user_key"] = identifier

        self.klasse_objects[identifier] = Klasse(**kwargs)

    def add_facet(self, identifier, **kwargs):
        """
        Add a Facet object to the map

        :param str identifier: A user defined identifier
        :param kwargs kwargs: :class:`os2mo_data_import.mox_data_types.Facet`
        """
        logger.warning("'add_facet' is deprecated. Use os2mo-init instead.")
        if identifier in self.facet_objects:
            raise ReferenceError("Unique constraint - Facet identifier exists")

        self.facet_objects[identifier] = Facet(**kwargs)

    def add_organisation_unit(self, identifier, **kwargs):
        """
        Add a OrganisationUnit object to the map

        :param str identifier: A user defined identifier
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.OrganisationUnitType`
        """

        if identifier in self.organisation_units:
            raise ReferenceError("Identifier exists")

        if "name" not in kwargs:
            kwargs["name"] = identifier

        self.organisation_units[identifier] = OrganisationUnitType(**kwargs)
        self.organisation_unit_details[identifier] = []

    def add_employee(self, identifier, **kwargs):
        """
        Add a Employee object to the map

        :param str identifier: A user defined identifier
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.EmployeeType`
        """

        if identifier in self.employees:
            raise ReferenceError("Identifier exists")

        kwargs['seperate_names'] = self.seperate_names

        self.employees[identifier] = EmployeeType(**kwargs)
        self.employee_details[identifier] = []

    def add_address_type(self, organisation_unit=None, employee=None, **kwargs):
        """
        Add a Address object to the map

        :param str organisation_unit: Reference to the parent unit
        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.AddressType`
        """

        if not (organisation_unit or employee):
            raise ReferenceError(
                "Either organisation unit or employee must be owner"
            )

        if organisation_unit and employee:
            raise ReferenceError(
                "Must reference either organisation unit or employee and not both"
            )

        if employee:

            if employee not in self.employees:
                raise ReferenceError("Owner does not exist")

            self.employee_details[employee].append(
                AddressType(**kwargs)
            )

        if organisation_unit:

            if organisation_unit not in self.organisation_units:
                raise ReferenceError("Owner does not exist")

            self.organisation_unit_details[organisation_unit].append(
                AddressType(**kwargs)
            )

    def add_engagement(self, employee, organisation_unit, **kwargs):
        """
        Add a Engagement object to the map

        :param str organisation_unit: Reference to the parent unit
        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.EngagementType`
        """

        engagement = EngagementType(org_unit_ref=organisation_unit, **kwargs)

        self.employee_details[employee].append(engagement)

    def add_association(self, employee, organisation_unit, **kwargs):
        """
        Add a Association object to the map

        :param str organisation_unit: Reference to the parent unit
        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.AssociationType`
        """

        association = AssociationType(org_unit_ref=organisation_unit, **kwargs)

        self.employee_details[employee].append(association)

    def terminate_employee(self, employee, **kwargs):
        """
        Terminate employee

        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.TerminationType`
        """
        termination = TerminationType(kwargs['date_from'])

        self.employee_details[employee].append(termination)

    def terminate_engagement(self, employee, **kwargs):
        """
        Terminate an engagement. Unfortunately it is only possibly to use
        a valdity of today.

        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.EngagementTerminationType`
        """
        termination = EngagementTerminationType(kwargs['engagement_uuid'])

        self.employee_details[employee].append(termination)

    def add_role(self, employee, organisation_unit, **kwargs):
        """
        Add a Role object to the map
        See available parameters for mora_data_types:RoleType

        :param str organisation_unit: Reference to the parent unit
        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.RoleType`
        """

        if employee not in self.employees:
            raise ReferenceError("Employee does not exist")

        if organisation_unit not in self.organisation_units:
            raise ReferenceError("Organisation unit does not exist")

        role = RoleType(org_unit=organisation_unit, **kwargs)

        self.employee_details[employee].append(role)

    def add_manager(self, employee, organisation_unit, **kwargs):
        """
        Add a Manager object to the map

        :param str organisation_unit: Reference to the parent unit
        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.ManagerType`
        """

        if employee not in self.employees:
            raise ReferenceError("Employee does not exist")

        if organisation_unit not in self.organisation_units:
            raise ReferenceError("Organisation unit does not exist")

        manager = ManagerType(org_unit=organisation_unit, **kwargs)

        self.employee_details[employee].append(manager)

    def add_leave(self, employee, **kwargs):
        """
        Add a Leave object to the map

        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.LeaveType`
        """

        if employee not in self.employees:
            raise ReferenceError("Employee does not exist")

        leave = LeaveType(**kwargs)

        self.employee_details[employee].append(leave)

    def new_itsystem(self, identifier, **kwargs):
        """
        Add a new Itsystem object to the map

        :param str identifier: A user defined identifier
        :param kwargs kwargs: :class:`os2mo_data_import.mox_data_types.Itsystem`
        """

        if identifier in self.itsystems:
            raise ReferenceError("It system already exists")

        self.itsystems[identifier] = Itsystem(**kwargs)

    def join_itsystem(self, employee, **kwargs):
        """
        Join Itsystem means adding a ItsystemType detail to the map

        :param str employee: Reference to the employee
        :param kwargs kwargs: :class:`os2mo_data_import.mora_data_types.ItsystemType`
        """
        if employee not in self.employees:
            raise ReferenceError("Employee does not exist")

        itsystem = ItsystemType(**kwargs)

        self.employee_details[employee].append(itsystem)

    def create_default_facet_types(self, facet_defaults=facet_defaults):
        """
        Add the default list of Facet type objects to the map
        For more information, please see the "facet_defaults" list
        of the defaults module.

        :param list facet_defaults: Default list of Facet types
        """
        logger.warning(
            "'create_default_facet_types' is deprecated. Use os2mo-init instead."
        )
        for user_key in facet_defaults:

            self.add_facet(
                identifier=user_key,
                user_key=user_key
            )

    def import_organisation_units_recursively(self, reference, org_unit):
        """
        Begin importing all the organisation unit objects
        from the map recursively.
        Organisation units without a parent reference
        automatically become "root" units.

        :param str reference: Reference to organisation_unit
        :param object org_unit: The OrganisationUnitType object
        """
        # Insert parents first!
        parent_ref = org_unit.parent_ref

        if parent_ref and parent_ref not in self.store.inserted_org_unit_map:
            parent_unit = self.organisation_units.get(parent_ref)

            # Insert parent first
            self.import_organisation_units_recursively(parent_ref, parent_unit)

        # Now insert actual units
        details = self.organisation_unit_details.get(reference)

        self.store.import_org_unit(
            reference=reference,
            organisation_unit=org_unit,
            details=details
        )

    def _import_unit_from_integration_data(self, reference):
        ou_res = 'organisation/organisationenhed'
        klasse_res = 'klassifikation/klasse'
        uuid = self.ia.find_object(ou_res, reference)
        # TODO: Should this include more than just present time?
        unit = self.morah.read_ou(uuid)

        type_uuid = unit['org_unit_type']['uuid']
        parent = unit['parent']
        date_from = unit['validity']['from']
        date_to = unit['validity']['to']
        type_ref = self.ia.read_integration_data(klasse_res, type_uuid)

        if parent:
            parent_uuid = parent['uuid']
            parent_ref = self.ia.read_integration_data(ou_res, parent_uuid)
        else:
            parent_ref = None

        self.add_organisation_unit(
            identifier=reference,
            parent_ref=parent_ref,
            type_ref=type_ref,
            uuid=unit['uuid'],
            date_from=date_from,
            date_to=date_to
        )

    def test_org_unit_refs(self, reference, org_unit):
        """
        Test if the parent ref to an org unit is already added to the importer map.
        If it is not, the parent will be imported to make the map self-consistent.
        :param reference: importer reference to the org unit to be tested.
        :param org_unit: Actual org unit object - used to identify the parent.
        :return: True if the parent was not found, indicating that the imporer
        map is not yet complete. False if the parent was already in place.
        """
        parent_unit = self.organisation_units.get(org_unit.parent_ref)

        if org_unit.parent_ref and (not parent_unit):
            re_run = True
            self._import_unit_from_integration_data(org_unit.parent_ref)
            logger.info('Importing {} from integration data'.format(org_unit.parent_ref))
        else:
            re_run = False
        return re_run

    def _import_organisation(self):
        self.store.import_organisation(*self.organisation)

    def _import_klassifikation(self):
        self.store.import_klassifikation(*self.klassifikation)

    def _import_facets(self):
        for identifier, facet in self.facet_objects.items():
            self.store.import_facet(identifier, facet)

    def _import_classes(self):
        for identifier, klasse in self.klasse_objects.items():
            self.store.import_klasse(identifier, klasse)

    def _import_itsystems(self):
        for identifier, itsystem in self.itsystems.items():
            self.store.import_itsystem(identifier, itsystem)

    def _import_org_units(self):
        re_run = True
        while re_run:
            re_run = False
            identifiers = list(self.organisation_units.keys())
            for identifier in identifiers:
                org_unit = self.organisation_units[identifier]
                # Test if the parent unit is in the map, if it is not, perform
                # an integration data based import from MO.
                # If the parent was not there, run once more to check if higher
                # levels of parents also needs to be imported.
                if self.test_org_unit_refs(identifier, org_unit):
                    re_run = True

        for identifier, org_unit in self.organisation_units.items():
            self.import_organisation_units_recursively(identifier, org_unit)

    def _import_employees(self):
        for identifier, employee in self.employees.items():
            details = self.employee_details.get(identifier)
            self.store.import_employee(
                reference=identifier,
                employee=employee,
                details=details
            )

    def import_all(self):
        """
        The import method begins importing all objects
        obtained from the maps in the following order:

            #. Organisation object
            #. Klassifikation object (auto)
            #. Facet objects
            #. Klasse objects

            #. OrganisationUnit objects
            #. Employees objects
        """

        # Insert Organisation
        logger.info('Will now import organisation')
        self._import_organisation()

        # Insert Klassifikation
        logger.info('Will now import klassifikation')
        self._import_klassifikation()

        # Insert Facet
        logger.info('Will now import facet')
        self._import_facets()

        # Insert Klasse
        logger.info('Will now import klasse')
        self._import_classes()

        # Insert Itsystem
        logger.info('Will now import IT-systems')
        self._import_itsystems()

        # Insert Organisation Units
        logger.info('Will now import org units')
        self._import_org_units()

        # Insert Employees
        logger.info('Will now import employees')
        self._import_employees()



import asyncio
from functools import partial
from itertools import product
from typing import Any, Sequence, Tuple
from uuid import UUID

import aiohttp
from jsonschema import validate

UUIDstr = str
aiosession = aiohttp.ClientSession


def is_uuid(string: str) -> bool:
    try:
        UUID(string)
        return True
    except ValueError:
        return False


def is_uuid_list(listy: Sequence[str]) -> bool:
    return all(map(is_uuid, listy))


def ensure_session(func):
    async def _decorator(self, *args, **kwargs):
        if self.session:
            return await func(self, session, *args, **kwargs)
        else:
            async with aiohttp.ClientSession() as session:
                return await func(self, session, *args, **kwargs)

    return _decorator


class ElementNotFound(Exception):
    pass


class MultipleElementsFound(Exception):
    pass


class MoxHelper:
    def __init__(self, hostname: str, session: aiosession = None):
        self.hostname = hostname
        self.session = session

    @ensure_session
    async def check_connection(self, session: aiosession) -> bool:
        url = self.hostname + "/version"
        try:
            async with session.get(url) as response:
                data = await response.json()
                if "lora_version" in data:
                    return True
        except Exception:
            pass
        return False

    @ensure_session
    async def generate_methods(self, session: aiosession) -> None:
        url = self.hostname + "/site-map"

        async def discover_endpoints():
            def build_service_tuple(url):
                # The URLs have the following format: /service/obj/fields
                leading_slash, service, obj, fields_suffix = url.split("/")
                return service, obj

            async with session.get(url) as response:
                data = await response.json()
                # Find all service urls in the site-map
                service_urls = filter(lambda x: x.endswith("/fields"), data["site-map"])
                return map(build_service_tuple, service_urls)

        service_tuples = list(await discover_endpoints())
        method_map = {
            "read_element": self._read_element,
            "read_all": self._read_uuid_list,
            "create": self._create,
            "get_or_create": self._get_or_create,
            "validate": self._validate_payload,
            "update": self._update,
            "search": self._search,
            "insert": self._insert,
        }
        # Generate each method for each service / object
        service_method_map = product(service_tuples, method_map.items())
        for (service, obj), (prefix, method) in service_method_map:
            method_name = "_".join([prefix, service, obj])
            setattr(self, method_name, partial(method, service, obj))
        # Fetch schemas for each endpoint in 'parallel'
        # TODO: Consider caching schemas on disk
        schema_tasks = map(
            lambda tup: asyncio.ensure_future(self._fetch_schema(*tup)), service_tuples
        )
        self.schemas = dict(await asyncio.gather(*schema_tasks))

    @ensure_session
    async def _fetch_schema(
        self, session: aiosession, service: str, obj: str
    ) -> Tuple[Tuple[str, str], str]:
        url = self.hostname + "/" + service + "/" + obj + "/schema"
        async with session.get(url) as response:
            data = await response.json()
            return (service, obj), data

    def _validate_payload(self, service: str, obj: str, payload: Any):
        schema = self.schemas[(service, obj)]
        validate(instance=payload, schema=schema)

    @ensure_session
    async def _search(
        self, session: aiosession, service: str, obj: str, params: Any
    ) -> Sequence[str]:
        url = self.hostname + "/" + service + "/" + obj
        async with session.get(url, params=params) as response:
            data = await response.json()
            return data["results"][0]

    @ensure_session
    async def _create(
        self, session: aiosession, service: str, obj: str, payload: Any
    ) -> UUIDstr:
        self._validate_payload(service, obj, payload)
        url = self.hostname + "/" + service + "/" + obj
        async with session.post(url, json=payload) as response:
            return (await response.json())["uuid"]

    @ensure_session
    async def _insert(
        self, session: aiosession, service: str, obj: str, payload: Any, uuid: UUIDstr
    ) -> UUIDstr:
        self._validate_payload(service, obj, payload)
        url = f"{self.hostname}/{service}/{obj}"
        if uuid:
            url = url + f"/{uuid}"
        async with session.put(url, json=payload) as response:
            return (await response.json())["uuid"]

    @ensure_session
    async def _update(
        self, session: aiosession, service: str, obj: str, uuid: UUIDstr, payload: Any
    ) -> UUIDstr:
        self._validate_payload(service, obj, payload)
        url = self.hostname + "/" + service + "/" + obj + "/" + uuid

        async with session.patch(url, json=payload) as response:
            return (await response.json())["uuid"]

    async def _read_uuid_list(self, service: str, obj: str) -> Sequence[UUIDstr]:
        result = await self._search(service, obj, {"bvn": "%"})
        # Check that we got back valid UUIDs
        if not is_uuid_list(result):
            raise ValueError("Endpoint did not return a list of uuids")
        return result

    async def _read_element(
        self, service: str, obj: str, params: Any = None, **kwargs
    ) -> UUIDstr:
        params = params or {}
        params.update(**kwargs)
        result = await self._search(service, obj, params)
        if not is_uuid_list(result):
            raise ValueError("Endpoint did not return a list of uuids")
        if len(result) == 1:
            return result[0]
        elif len(result) > 1:
            raise MultipleElementsFound
        elif len(result) == 0:
            raise ElementNotFound

    async def _get_or_create(
        self, service: str, obj: str, payload, **kwargs
    ) -> UUIDstr:
        bvn = payload["attributter"][obj + "egenskaber"][0]["brugervendtnoegle"]
        params = {"bvn": bvn}
        params.update(**kwargs)
        try:
            element = await self._read_element(service, obj, params=params)
            return element, False
        except ElementNotFound:
            element = await self._create(service, obj, payload)
            return element, True


async def create_mox_helper(*args, generate_methods=True, **kwargs):
    mox_helper = MoxHelper(*args, **kwargs)
    if generate_methods:
        await mox_helper.generate_methods()
    return mox_helper


if __name__ == "__main__":

    async def run():
        mox = await create_mox_helper("http://localhost:8080")
        print(await mox.check_connection())

    loop = asyncio.new_event_loop()
    loop.run_until_complete(run())



def _virkning(dato='1910-01-01 00:00:00'):
    virkning = {
        'from': dato,
        'to': 'infinity',
        'aktoerref': 'ddc99abd-c1b0-48c2-aef7-74fea841adae',
        'aktoertypekode': 'Bruger'
    }
    return virkning


def lora_facet(bvn, org='ddc99abd-c1b0-48c2-aef7-74fea841adae'):
    attributter = {
        'facetegenskaber': [{
            'brugervendtnoegle': bvn,
            'virkning': _virkning(),
        }]
    }
    tilstande = {
        'facetpubliceret': [{
            'publiceret': 'Publiceret',
            'virkning': _virkning()
        }]
    }
    relationer = {
        "ansvarlig": [{
            'objekttype': 'organisation',
            'uuid': org,
            'virkning': _virkning()
        }]
    }
    facet = {
        'attributter': attributter,
        'tilstande': tilstande,
        'relationer': relationer
    }
    return facet


def lora_klasse(brugervendtnoegle, beskrivelse, titel, dato, facet, ansvarlig, omfang=None,
                overklasse=None):
    attributter = {
        'klasseegenskaber': [{
            'brugervendtnoegle': brugervendtnoegle,
            'beskrivelse': beskrivelse,
            'titel': titel,
            'virkning': _virkning(dato)
        }]
    }
    tilstande = {
        'klassepubliceret': [{
            'publiceret': 'Publiceret',
            'virkning': _virkning(dato)
        }]
    }
    relationer = {
        'facet': [{
            'uuid': facet,
            'virkning': _virkning(dato),
            'objekttype': 'Facet'
        }],
        'overordnetklasse': [{
            'virkning': _virkning(dato),
            'objekttype': 'Klasse'
        }],
        'ansvarlig': [{
            'uuid': ansvarlig,
            'virkning': _virkning(dato),
            'objekttype': 'Organisation'
        }]
    }
    klasse = {
        'attributter': attributter,
        'tilstande': tilstande,
        'relationer': relationer
    }
    if overklasse is not None:
        klasse['relationer']['overordnetklasse'][0]['uuid'] = overklasse
    else:
        del klasse['relationer']['overordnetklasse']

    if omfang:
        attributter['klasseegenskaber'][0]['omfang'] = omfang

    return klasse



import uuid


def _virkning(dato=None):
    if dato is None:
        dato = "1910-01-01 00:00:00"
    virkning = {
        "from": dato,
        "to": "infinity",
        "aktoerref": "ddc99abd-c1b0-48c2-aef7-74fea841adae",
        "aktoertypekode": "Bruger",
    }
    return virkning


def lora_facet(bvn, org_uuid, description=None):

    attributter = {
        "facetegenskaber": [{"brugervendtnoegle": bvn, "virkning": _virkning(),}]
    }
    if description:
        attributter["facetegenskaber"][0]["beskrivelse"] = description

    relationer = {
        "ansvarlig": [
            {"objekttype": "organisation", "uuid": org_uuid, "virkning": _virkning()}
        ]
    }
    tilstande = {
        "facetpubliceret": [{"publiceret": "Publiceret", "virkning": _virkning()}]
    }
    facet = {
        "attributter": attributter,
        "relationer": relationer,
        "tilstande": tilstande,
    }
    return facet


def lora_klasse(
    bvn,
    title,
    facet_uuid,
    org_uuid,
    org_unit_uuid=None,
    description=None,
    dato=None,
    scope=None,
    parent_uuid=None,
):
    # NOTE: This is used from SD, and should be split out as a library?
    attributter = {
        "klasseegenskaber": [
            {
                "brugervendtnoegle": bvn,
                "titel": title,
                "virkning": _virkning(dato),
            }
        ]
    }
    if description:
        attributter["klasseegenskaber"][0]["beskrivelse"] = description
    if scope:
        attributter["klasseegenskaber"][0]["omfang"] = scope
    tilstande = {
        "klassepubliceret": [
            {"publiceret": "Publiceret", "virkning": _virkning(dato)}
        ]
    }
    relationer = {
        "facet": [
            {"uuid": facet_uuid, "virkning": _virkning(dato), "objekttype": "Facet"}
        ],
        "overordnetklasse": [
            {"uuid": parent_uuid, "virkning": _virkning(dato), "objekttype": "Klasse"}
        ],
        "ansvarlig": [
            {
                "uuid": org_uuid,
                "virkning": _virkning(dato),
                "objekttype": "Organisation",
            }
        ],
        "ejer": [
            {
                "uuid": org_unit_uuid,
                "virkning": _virkning(dato),
                "objekttype": "OrganisationEnhed",
            }
        ],
    }
    if parent_uuid is None:
        del relationer["overordnetklasse"]
    if org_unit_uuid is None:
        del relationer["ejer"]
    klasse = {
        "attributter": attributter,
        "relationer": relationer,
        "tilstande": tilstande,
    }

    return klasse



from unittest import TestCase
from unittest.mock import patch

from parameterized import parameterized
from requests.exceptions import ConnectionError
from requests.exceptions import HTTPError

from ..mora_helpers import MoraHelper


class TestMoraHelper(TestCase):
    def setUp(self):
        super().setUp()
        self._instance = MoraHelper()

    @parameterized.expand(
        [
            (ConnectionError, 5),
            (HTTPError, 5),
            (Exception, 1),
        ]
    )
    def test_mo_lookup_retries_on_error(self, exception_class, expected_attempts):
        """Check that the expected number of retry attempts are performed, depending
        on the specific exception raised by `requests.get`.
        """
        with self.assertRaises(exception_class):
            with patch("requests.get", side_effect=exception_class):
                self._instance._mo_lookup("not-a-uuid", "not-a-url")
                self.assertEqual(
                    self._instance._mo_lookup.retry.statistics["attempt_number"],
                    expected_attempts,
                )



import json
import logging
import pathlib
import requests
import xmltodict

from kle import payloads

LOG_FILE = 'kle_online.log'

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    filename=LOG_FILE
)

# Facetter
# Emne: http://api.kle-online.dk/resources/kle/emneplan
# Funktion: http://api.kle-online.dk/resources/kle/handlingsfacetter


class KleImporter(object):
    """ Script to import KLE into LoRa to allow easy access to relevant
    test data

    In short, KLE consists of four nested levels, one Facet level and up to
    three Klasse levels. The amount of code-reuse between the levels is fairly
    small since the details in the XML files in the three levels are
    somewhat different. The task of actually retriveing the information
    about each class is put in seperate functions, read_all_from_, in
    order to keep flexibility if it becomes relevant to extend the amount
    of imported information for some of the levels. Possibly the current
    genereic json-template will be too simple in this case, it might be
    necessary with seperate templates for the various levels.
    """

    def __init__(self, mox_base, mora_base, api_token=None):
        """
        Init function
        :para hostname: hostname for the rest interface
        """
        self.mox_base = mox_base
        self.mora_base = mora_base
        self.mo_session = requests.Session()
        if api_token:
            self.mo_session.headers = {"SESSION": api_token}

    def _read_kle_dict(self, facet='emne', local=False):
        """ Read the entire KLE file
        :param facet: Either 'emne' or 'handlingsfacetter'
        :param local: If True the file is read from local cache
        :return: The document date and the KLE index as a dict
        """
        if facet == 'emne':
            navn = 'emneplan'
        else:
            navn = 'handlingsfacetter'

        if local:
            with open('integrations/kle/' + navn + '.xml', 'r') as content_file:
                xml_content = content_file.read()
        else:
            url = 'http://api.kle-online.dk/resources/kle/'
            response = requests.get(url + navn)
            response.encoding = 'utf-8'
            xml_content = response.text

        kle_dict = xmltodict.parse(xml_content)
        if facet == 'emne':
            udgivelses_dato = kle_dict['KLE-Emneplan']['UdgivelsesDato']
            kle_dict = kle_dict['KLE-Emneplan']['Hovedgruppe']
        else:
            udgivelses_dato = kle_dict['KLE-Handlingsfacetter']['UdgivelsesDato']
            kle_dict = kle_dict['KLE-Handlingsfacetter']['HandlingsfacetKategori']
        return (udgivelses_dato, kle_dict)

    def get_or_create_facet(self, facet_name):
        """
        Creates a new facet or returns an existing one
        :param facet_name: Name of the new facet
        :return: Returns uuid of the new facet
        """
        url = '/klassifikation/facet'

        loraurl = (
            self.mox_base +
            '/klassifikation/facet' +
            "?bvn=" + facet_name
        )

        lora_all = requests.get(loraurl).json()["results"][0]

        if len(lora_all) == 1:
            return lora_all[0]

        elif len(lora_all) == 0:
            template = payloads.lora_facet(bvn=facet_name, org=self.org_uuid)
            logger.info("creating facet: %r", template)
            response = requests.post(self.mox_base + '/klassifikation/facet',
                                     json=template)
            return response.json()['uuid']

        else:
            logger.error("Mere end en instans af facetten '%s' fundet" % facet_name)
            raise RuntimeError("Facet Dublet: %s" % facet_name)


    def get_or_create_klasse(self, facet, klasse_info, overklasse=None):
        """
        Creates a new Klasse based on KLE
        :param facet: uuid for the korresponding facet
        :param klasse_info: Dict as returned by read_all_*
        :return: Returns uuid of the new klasse
        """
        uuid = klasse_info['uuid']

        payload = payloads.lora_klasse(
            brugervendtnoegle=klasse_info['nummer'],
            beskrivelse=klasse_info['titel'],
            titel="{} - {}".format(klasse_info['nummer'], klasse_info['titel']),
            facet=facet,
            dato=klasse_info['dato'],
            overklasse=overklasse,
            ansvarlig=self.org_uuid)

        return self.get_or_create_lora_klasse(payload, uuid)

    def get_or_create_lora_klasse(self, payload, uuid):
        url = '/klassifikation/klasse/{}'
        full_url = self.mox_base + url.format(uuid)

        response = requests.get(full_url)
        if response.status_code not in [200, 404]:
            logger.error("Loraopslag fejlede")
            response.raise_for_status()

        if response.status_code == 404:
            logger.info("creating lora klasse: %r on %s", payload, uuid)
            response = requests.put(full_url, json=payload)
            lora_uuid = response.json()['uuid']
            assert lora_uuid == uuid

        return uuid

    def _read_all_hovedgrupper(self, facet='emne'):
        """ Read all Hovedgrupper from KLE
        :param facet: Either 'emne' or 'handlingsfacetter'
        :return: Dict with index as as key and
        (HovedgruppeTitel, HovedgruppeNr) as value
        """
        name = 'Hovedgruppe' if facet == 'emne' else 'HandlingsfacetKategori'
        hovedgrupper = {}
        for i in range(0, len(self.kle_dict)):
            titel = self.kle_dict[i][name + 'Titel']
            hovedgrupper[i] = (titel, self.kle_dict[i][name + 'Nr'])
        return hovedgrupper

    def _read_all_from_hovedgruppe(self, hovedgruppe_index, facet='emne'):
        """
        Read all relevant fields from a Hovedgruppe - this can
        easily be extended if more info turns out to be relevant
        :param hovedgruppe_index: Index for the wanted Hovedgruppe
        :param facet: Either 'emne' or 'handlingsfacetter'
        :return: Dict with relevant info
        """
        name = 'Hovedgruppe' if facet == 'emne' else 'HandlingsfacetKategori'
        hovedgruppe = self.kle_dict[hovedgruppe_index]
        hovedgruppe_info = {
            'uuid': hovedgruppe['UUID'],
            'titel': hovedgruppe[name + 'Titel'],
            'dato': hovedgruppe[name + 'AdministrativInfo']['OprettetDato'],
            'nummer': hovedgruppe[name + 'Nr']
        }
        # TODO: Der findes også info om rettet-dato, er dette relevant?
        return hovedgruppe_info

    def _read_all_grupper(self, hovedgruppe, facet='emne'):
        """ Read all Grupper from a KLE Hovedgruppe
        :param hovedgruppe: A KLE Hovedgruppe index to be retrieved
        :param facet: Either 'emne' or 'handlingsfacetter'
        :return: Dict with index as key and (GruppeTitel, GruppeNr) as value
        """
        name = 'Gruppe' if facet == 'emne' else 'Handlingsfacet'
        grupper = {}
        gruppe_liste = self.kle_dict[hovedgruppe][name]
        for i in range(0, len(gruppe_liste)):
            grupper[i] = (gruppe_liste[i][name + 'Titel'],
                          gruppe_liste[i][name + 'Nr'])
        return grupper

    def _read_all_from_gruppe(self, hovedgruppe, gruppe, facet='emne'):
        """ Read all relevant fields from a Gruppe - this can
        easily be extended if more info turns out to be relevant
        :param hovedgruppe: Index for the wanted Hovedgruppe
        :param gruppe: Index for the wanted Gruppe
        :param facet: Either 'emne' or 'handlingsfacetter'
        :return: Dict with relevant info
        """
        name = 'Gruppe' if facet == 'emne' else 'Handlingsfacet'
        gruppe = self.kle_dict[hovedgruppe][name][gruppe]

        gruppe_info = {
            'uuid': gruppe['UUID'],
            'titel': gruppe[name + 'Titel'],
            'dato': gruppe[name + 'AdministrativInfo']['OprettetDato'],
            'nummer': gruppe[name + 'Nr'],
        }
        return gruppe_info

    def _read_all_emner(self, hovedgruppe, gruppe):
        """ Read all Emner from a KLE Gruppe
        :param hovedgruppe: The KLE Hovedgruppe index containing the Gruppe
        :return: Dict with index as key and (EmneTitel, EmneNr) as value
        """
        emner = {}
        emne_liste = self.kle_dict[hovedgruppe]['Gruppe'][gruppe]['Emne']
        for i in range(0, len(emne_liste)):
            try:
                emner[i] = (emne_liste[i]['EmneTitel'],
                            emne_liste[i]['EmneNr'][6:])
            except KeyError:  # If only one element, there is no list
                emner[0] = (emne_liste['EmneTitel'],
                            emne_liste['EmneNr'])
        return emner

    def _read_all_from_emne(self, hovedgruppe, gruppe, emne):
        """ Read all relevant fields from a Gruppe - this can
        easily be extended if more info turns out to be relevant
        :param hovedgruppe: Index for the wanted Hovedgruppe
        :param gruppe: Index for the wanted Gruppe
        :param gruppe: emne for the wanted Emne
        :return: Dict with relevant info
        """
        try:
            emne = self.kle_dict[hovedgruppe]['Gruppe'][gruppe]['Emne'][emne]
        except KeyError:  # If only one element, there is no list
            emne = self.kle_dict[hovedgruppe]['Gruppe'][gruppe]['Emne']

        emne_info = {
            'uuid': emne['UUID'],
            'titel': emne['EmneTitel'],
            'dato': emne['EmneAdministrativInfo']['OprettetDato'],
            'nummer': emne['EmneNr']
        }
        return emne_info

    def _import_emne(self, facet_uuid):
        """
        Read all classes from the KLE facet 'emneplan'
        :param facet_uuid: UUID of the facet the classes are created under
        """
        logger.info("Importing 'Emne'")
        kle_content = self._read_kle_dict(facet='emne')
        self.kle_dict = kle_content[1]

        hovedgrupper = self._read_all_hovedgrupper()
        for hoved_index in hovedgrupper:
            hoved_info = self._read_all_from_hovedgruppe(hoved_index)
            # Create hovedgruppe
            hoved_uuid = self.get_or_create_klasse(facet_uuid, hoved_info)

            grupper = self._read_all_grupper(hoved_index)
            for gruppe_index in grupper:
                gruppe_info = self._read_all_from_gruppe(hoved_index, gruppe_index)
                # Create gruppe
                gruppe_uuid = self.get_or_create_klasse(facet_uuid, gruppe_info,
                                                        hoved_uuid)
                emner = self._read_all_emner(hoved_index, gruppe_index)
                for emne_index in emner:
                    emne_info = self._read_all_from_emne(hoved_index,
                                                         gruppe_index,
                                                         emne_index)
                    # Create emne
                    self.get_or_create_klasse(facet_uuid, emne_info, gruppe_uuid)

    def _import_handling(self, facet_uuid):
        """
        Import the classes from the KLE facet 'handlingsfacetter'
        :param facet_uuid: UUID of the facet the classes are created under
        """
        logger.info("Importing 'Handling'")
        facet = 'handling'
        kle_content = self._read_kle_dict(facet=facet)
        self.kle_dict = kle_content[1]

        hovedgrupper = (self._read_all_hovedgrupper(facet=facet))
        for hoved_index in hovedgrupper:
            hoved_info = self._read_all_from_hovedgruppe(hoved_index,
                                                         facet=facet)
            hoved_uuid = self.get_or_create_klasse(facet_uuid, hoved_info)

            grupper = self._read_all_grupper(hoved_index, facet=facet)
            for gruppe_index in grupper:
                gruppe_info = self._read_all_from_gruppe(hoved_index,
                                                         gruppe_index,
                                                         facet=facet)
                # Create gruppe
                self.get_or_create_klasse(facet_uuid, gruppe_info, hoved_uuid)

    def set_mo_org_uuid(self):
        mora_base = self.mora_base
        r = self.mo_session.get("{}/service/o/".format(mora_base))
        r.raise_for_status()
        self.org_uuid = r.json()[0]['uuid']

    def import_aspect_classes(self, facet_uuid):
        logger.info('Importing aspect classes')
        for key, scope, uuid in [
            ('Indsigt', 'INDSIGT', '92f719f2-9e34-459e-8610-8dd160747a93'),
            ('Udførende', 'UDFOERENDE', '40e91f7a-d2fc-4e07-8108-4046bde113d0'),
            ('Ansvarlig', 'ANSVARLIG', '4d9d0ff4-017d-4e34-acc4-d403f7b2358c')
        ]:
            payload = payloads.lora_klasse(
                brugervendtnoegle=key,
                beskrivelse=key,
                titel=key,
                facet=facet_uuid,
                omfang=scope,
                dato="1910-01-01 00:00:00",
                ansvarlig=self.org_uuid)
            self.get_or_create_lora_klasse(payload, uuid)

    def _show_kle(self):
        """Configure MO to show KLE."""
        mora_base = self.mora_base
        payload={"org_units": {"show_kle": True}}
        r = self.mo_session.post("{}/service/configuration".format(mora_base), json=payload)
        r.raise_for_status()

    def import_kle(self):
        self.set_mo_org_uuid()
        aspect_facet_uuid = self.get_or_create_facet('kle_aspect')
        self.import_aspect_classes(aspect_facet_uuid)
        number_facet_uuid = self.get_or_create_facet('kle_number')
        self._import_emne(number_facet_uuid)
        self._import_handling(number_facet_uuid)
        self._show_kle()


if __name__ == '__main__':
    cfg_file = pathlib.Path.cwd() / 'settings' / 'settings.json'
    if not cfg_file.is_file():
        raise Exception('No setting file')
    settings = json.loads(cfg_file.read_text())

    mora_base = settings['mora.base']
    mox_base = settings['mox.base']
    api_token = settings.get('crontab.SAML_TOKEN')

    kle = KleImporter(mox_base, mora_base, api_token)
    kle.import_kle()
    logger.info("program has ended")



##
# Copyright (c) 2017-2018, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
"""
Helper class to make a number of pre-defined queries into MO
"""
import codecs
import csv
import datetime
import logging
from functools import lru_cache
from operator import itemgetter
from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

import requests
from anytree import Node
from more_itertools import one
from more_itertools import only
from ra_utils.headers import TokenSettings
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import wait_random_exponential
from tenacity import stop_after_attempt

PRIMARY_RESPONSIBILITY = "Personale: ansættelse/afskedigelse"

logger = logging.getLogger("mora-helper")


class MoraHelper:
    def __init__(
        self, hostname="http://localhost:5000", export_ansi=True, use_cache=True
    ):
        self.host = hostname + "/service/"
        self.cache = {}
        self.default_cache = use_cache
        self.export_ansi = export_ansi

    def _split_name(self, name):
        """Split a name into first and last name.
        Currently just splits at last space, but this might turn out to
        be a too simple rule.
        :param name: The name to split.
        :return: Dict with first and last name separated.
        """
        splitted_name = {
            "Fornavn": name[: name.rfind(" ")],
            "Efternavn": name[name.rfind(" ") :],
        }
        return splitted_name

    def _read_node_path(self, node):
        """Find the full path for a given node
        :param node: The node to find the path for
        :return: List with full path for the node
        """
        path = []
        for sub_node in node.path:
            ou = self.read_ou(sub_node.name)
            path += [ou["name"]]
        return path

    def _create_fieldnames(self, nodes):
        """Create a list of fieldnames that has a suitable length
        to accomodate the full tree debth. First names are hard-coded
        the rest are auto-generated
        :param nodes: List of all nodes in tree
        :return: A list of fieldnames
        """
        fieldnames = ["root", "org", "sub org"]
        for i in range(2, nodes["root"].height):
            fieldnames += [str(i) + "xsub org"]
        return fieldnames

    def _write_csv(self, fieldnames, rows, filename):
        """Write a csv-file from a a dataset. Only fields explicitly mentioned
        in fieldnames will be saved, the rest will be ignored.
        :param fieldnames: The headline for the columns, also act as filter.
        :param rows: A list of dicts, each list element will become a row, keys
        in dict will be matched to fieldnames.
        :param filename: The name of the exported file.
        """
        print("Encode ascii: {}".format(self.export_ansi))
        with open(filename, encoding="utf-8", mode="w") as csvfile:
            writer = csv.DictWriter(
                csvfile,
                fieldnames=fieldnames,
                extrasaction="ignore",
                delimiter=";",
                quoting=csv.QUOTE_ALL,
            )
            writer.writeheader()
            for row in rows:
                writer.writerow(row)

        if self.export_ansi:
            with codecs.open(filename, "r", encoding="utf-8") as csvfile:
                lines = csvfile.read()
            with codecs.open(filename, "w", encoding="cp1252") as csvfile:
                csvfile.write(lines)

    def _create_path_dict(self, fieldnames, node, org_types=None):
        """Create a dict with a MO-path to a given node.

        :param fieldnames: The headline for each level of the path.
        :node: The node to find the path for.
        :return: A dict with headlines as keys and nodes as values.
        """
        ou = self.read_ou(node.name)
        ou_type = ou["org_unit_type"]["name"]

        if org_types and (ou_type not in org_types):
            return None

        path = self._read_node_path(node)
        path_dict = {}
        i = 0
        for element in path:
            path_dict[fieldnames[i]] = element
            i += 1
        return path_dict

    @retry(
        retry=retry_if_exception_type(
            (requests.exceptions.HTTPError, requests.exceptions.ConnectionError)
        ),
        wait=wait_random_exponential(multiplier=1, max=5),
        stop=stop_after_attempt(5),
        reraise=True,
    )
    def _mo_lookup(
        self,
        uuid,
        url,
        at=None,
        validity=None,
        only_primary=False,
        use_cache=None,
        calculate_primary=False,
    ):
        # TODO: at-value is currently not part of cache key
        if use_cache is None:
            use_cache = self.default_cache

        params = {}
        if calculate_primary:
            params["calculate_primary"] = 1
        if only_primary:
            params["only_primary_uuid"] = 1
        if at:
            params["at"] = at
        elif validity:
            params["validity"] = validity

        full_url = self.host + url.format(uuid)
        cache_id = full_url + str(validity)
        if (cache_id in self.cache) and use_cache:
            logger.debug("cache hit: %s", cache_id)
            return_dict = self.cache[cache_id]
        else:
            headers = TokenSettings().get_headers()

            response = requests.get(full_url, headers=headers, params=params)
            if response.status_code == 401:
                msg = "Missing Authorization"
                if headers:
                    msg = "Authorization not accepted"
                logger.error(msg)
                raise requests.exceptions.RequestException(msg)

            if (response.status_code == 500) and (
                "has been deleted" not in response.text
            ):
                response.raise_for_status()

            return_dict = response.json()
            self.cache[cache_id] = return_dict

        return return_dict

    def _mo_post(self, url, payload, force=True):
        params = {}
        if force:
            params["force"] = 1

        headers = TokenSettings().get_headers()

        full_url = self.host + url
        response = requests.post(full_url, headers=headers, params=params, json=payload)
        return response

    def check_connection(self):
        """Check that a connection can be established to MO."""
        # Really any endpoint could be used here
        response = self._mo_lookup(uuid=None, url="configuration")
        return "show_roles" in response

    @lru_cache
    def read_organisation(self):
        """Read the main Organisation, all OU's will have this as root

        :return: UUID of root organisation
        """
        org_id = self._mo_lookup(uuid=None, url="o/")
        return org_id[0]["uuid"]

    def read_all_users(self, limit: int = 1_000, at=None) -> List[dict]:
        """Return a list of all employees in MO.

        :param limit: Size of pagination groups. Set to 0 to skip pagination and fetch all users in one request.
        :return: List af all employees.
        """
        org = self.read_organisation()
        start = 0
        total = 1
        all_employees = []
        while start < total:
            employee_list = self._mo_lookup(org, f"o/{{}}/e/?limit={limit}&start={start}", at=at)
            all_employees.extend(employee_list["items"])
            start = employee_list['offset'] + limit
            total = employee_list['total']
        return all_employees


    def read_it_systems(self):
        """Read all it systems

        :return: list of it systems, eg. Opus, Active Directory
        """
        org_id = self.read_organisation()
        it_systems = self._mo_lookup(org_id, url="o/{}/it/")
        return it_systems

    def read_ou(self, uuid, at=None, use_cache=None) -> Dict:
        """Return a dict with the data available about an OU.

        :param uuid: The UUID of the OU
        :return: Dict with the information about the OU
        """
        org_enhed = self._mo_lookup(uuid, "ou/{}/", at, use_cache)
        return org_enhed

    def read_ou_address(
        self,
        uuid,
        at=None,
        use_cache=None,
        scope="DAR",
        return_all=False,
        reformat=True,
    ):
        """Return a dict with the data available about an OU.

        :param uuid: The UUID of the OU
        :param return_all: If True the response will be a list of dicts
        rather than a dict, and all adresses will be returned.
        :return: Dict (or list) with the information about the OU
        """

        def reformat_address(address):
            return {
                "type": address["address_type"]["uuid"],
                "visibility": address.get("visibility"),
                # Deprecated spelling mistake....
                "visibibility": address.get("visibility"),
                "Adresse": address["name"],
                "value": address["value"],
                "uuid": address["uuid"],
            }

        addresses = self._mo_lookup(uuid, "ou/{}/details/address", at, use_cache)
        if scope is not None:
            addresses = filter(
                lambda address: address["address_type"]["scope"] == scope, addresses
            )
        # TODO: Eliminiate this random intermediate format?
        if reformat:
            addresses = map(reformat_address, addresses)
        return_list = list(addresses)

        if return_all:
            return return_list
        if return_list:
            return return_list[0]
        return {}

    def read_classes_in_facet(self, facet, use_cache=False):
        """Return all classes belong to a given facet.

        :param facet: The facet to be returned.
        :return: List of classes in the facet and the uuid of the facet.
        """
        org_uuid = self.read_organisation()
        url = "o/" + org_uuid + "/f/{}/"
        class_list = self._mo_lookup(facet, url, use_cache=False)
        classes = class_list["data"]["items"]
        facet_uuid = class_list["uuid"]
        return (classes, facet_uuid)

    @lru_cache
    def ensure_class_in_facet(
        self, facet: str, bvn: str, title: Optional[str] = None, scope: str = "TEXT", uuid: Optional[UUID] = None
    ) -> UUID:
        """Ensures class exists in given facet."""

        classes, _ = self.read_classes_in_facet(facet)
        current = list(filter(lambda c: c["user_key"] == bvn, classes))

        if current:
            return UUID(one(current)["uuid"])

        org_uuid = self.read_organisation()
        title = title or bvn
        payload = {"name": title, "user_key": bvn, "scope": scope, "org_uuid": org_uuid}
        url = f"f/{facet}/"
        if uuid:
            payload.update({"uuid": str(uuid)})
        r = self._mo_post(url, payload)
        r.raise_for_status()
        return UUID(r.json())

    def read_user(
        self, user_uuid=None, user_cpr=None, at=None, use_cache=None, org_uuid=None
    ):
        """Read basic info for a user. Either uuid or cpr must be given.

        :param user_uuid: UUID of the wanted user.
        :param user_cpr: cpr of the wanted user.
        :return: Basic user info
        """
        user_info = None
        if user_uuid:
            user_info = self._mo_lookup(user_uuid, "e/{}/", at, use_cache)
        if user_cpr:
            if not org_uuid:
                org_uuid = self.read_organisation()
            user = self._mo_lookup(
                user_cpr, "o/" + org_uuid + "/e/?query={}", at, use_cache
            )
            assert user["total"] < 2  # Only a single person can be found from cpr

            if user["total"] == 1:
                user_info = self._mo_lookup(
                    user["items"][0]["uuid"], "e/{}/", at, use_cache
                )
        return user_info

    def terminate_detail(self, mo_type, uuid, from_date):
        """Terminate a specific MO detail.

        :param mo_type: The MO type to terminate (association, role, engagement, etc)
        :param uuid: Object to terminate.
        :param from_date: Date to terminate from.
        """
        payload = {"type": mo_type, "uuid": uuid, "validity": {"to": "2018-01-01"}}
        logger.info("Terminate detail %s", payload)
        # self._mo_post('details/terminate', payload):

    def read_user_engagement(self, *args, **kwargs):
        return self.read_user_engagements(*args, **kwargs)

    def read_user_engagements(
        self,
        user,
        at=None,
        read_all=False,
        skip_past=False,
        only_primary=False,
        use_cache=None,
        calculate_primary=False,
    ):
        """Read engagements for a user.

        :param user: UUID of the wanted user.
        :read_all: Read all engagements, not only the present ones.
        :skip_past: Even if read_all is true, do not read the past.
        :calculate_primary: If True, ask MO to calculate primary engagement status.
        :return: List of the users engagements.
        """
        if not read_all:
            engagements = self._mo_lookup(
                user,
                "e/{}/details/engagement",
                at,
                only_primary=only_primary,
                use_cache=use_cache,
                calculate_primary=calculate_primary,
            )
        else:
            if skip_past:
                validity_times = ["present", "future"]
            else:
                validity_times = ["past", "present", "future"]

            engagements = []
            for validity in validity_times:
                engagement = self._mo_lookup(
                    user,
                    "e/{}/details/engagement",
                    validity=validity,
                    only_primary=only_primary,
                    use_cache=False,
                    calculate_primary=calculate_primary,
                )
                if isinstance(engagement, list):
                    engagements = engagements + engagement
                else:
                    print(f"There was an unexpected engagement returned: {engagement}")
        return engagements

    def read_user_association(
        self, user, at=None, read_all=False, only_primary=False, use_cache=None
    ):
        """Read associations for a user.

        :param user: UUID of the wanted user.
        :return: List of the users associations.
        """
        if not read_all:
            associations = self._mo_lookup(
                user,
                "e/{}/details/association",
                at,
                only_primary=only_primary,
                use_cach=use_cache,
            )
        else:
            associations = []
            for validity in ["past", "present", "future"]:
                association = self._mo_lookup(
                    user,
                    "e/{}/details/association",
                    validity=validity,
                    only_primary=only_primary,
                    use_cache=False,
                )
                associations = associations + association
        return associations

    def read_user_address(
        self,
        user,
        username=False,
        cpr=False,
        at=None,
        use_cache=None,
        phone_type=None,
        email_type=None,
    ):
        """Read phone number and email from user.

        :param user: UUID of the wanted user
        :param phone_type: Optionally add a specific phone_type class.
        :return: Dict witn phone number and email (if they exists in MO)
        """
        addresses = self._mo_lookup(user, "e/{}/details/address", at, use_cache)
        return_address = {}
        for address in addresses:
            if address["address_type"]["scope"] == "PHONE":
                if phone_type is None:
                    return_address["Telefon"] = address["name"]
                else:
                    if address["address_type"]["uuid"] == phone_type:
                        return_address["Telefon"] = address["name"]

            if address["address_type"]["scope"] == "EMAIL":
                if email_type is None:
                    return_address["E-mail"] = address["name"]
                else:
                    if address["address_type"]["uuid"] == email_type:
                        return_address["E-mail"] = address["name"]
        if username or cpr:
            personal_info = self._mo_lookup(user, "e/{}/")
            if username:
                return_address["Brugernavn"] = personal_info["user_key"]
            if cpr:
                return_address["CPR-Nummer"] = personal_info["cpr_no"]
        return return_address

    def read_user_manager_status(self, user):
        """Returns True of user has responsibility:
        'Personale: ansættelse/afskedigelse'
        :param user: UUID of the wanted user
        :return: True if person is manager accoring to above mentioned rule
        """
        manager_functions = self._mo_lookup(user, "e/{}/details/manager")
        for manager_function in manager_functions:
            for responsibility in manager_function["responsibility"]:
                if responsibility["name"] == PRIMARY_RESPONSIBILITY:
                    return True
        return False

    def read_user_roller(self, user):
        """Returns the role of the user
        :param user: UUID of the wanted user
        :return: The roles for the user
        """
        roles = self._mo_lookup(user, "e/{}/details/role")
        role_types = []
        for role in roles:
            role_types.append(role["role_type"]["name"])
        return role_types

    def read_engagement_manager(self, engagement_uuid):
        """
        Read the manager corresponding to a given engagement.
        If the engagement user is not a manager, the manager will be the (possibly
        inhereted) manager of the unit. If the person is a manager, the manager will
        be the manager of the nearest higer level department. If no higher level
        managers exists, the person will be a self-manager (typically
        'Kommunaldirektør').
        """
        user_manager = None

        # XXX: Why is this here, this is for MOX?!
        url = "http://localhost:8080/organisation/organisationfunktion/{}"
        response = requests.get(url.format(engagement_uuid))
        data = response.json()
        relationer = data[engagement_uuid][0]["registreringer"][0]["relationer"]
        user = relationer["tilknyttedebrugere"][0]
        unit = relationer["tilknyttedeenheder"][0]

        unit_manager = self.read_ou_manager(unit["uuid"], inherit=True)
        if unit_manager is None:
            raise Exception("Unable to find manager")

        if unit_manager["uuid"] != user["uuid"]:
            # In this case the engagement is not manager for itself
            user_manager = unit_manager
        else:
            mo_unit = self.read_ou(unit["uuid"])
            while unit_manager["uuid"] == user["uuid"]:
                if mo_unit["parent"] is None:
                    # Self manager!
                    break
                parent_uuid = mo_unit["parent"]["uuid"]
                unit_manager = self.read_ou_manager(parent_uuid, inherit=True)
                mo_unit = self.read_ou(parent_uuid)
            user_manager = unit_manager

        return user_manager

    def read_ou_manager(self, unit_uuid, inherit=False):
        """
        Read the manager of an organisation.
        Currently an exception will be raised if an ou has more than
        one manager.
        :param org_uuid: UUID of the OU to find the manager of
        :return: Returns 0 or 1 manager of the OU.
        """
        manager_list = {}
        if inherit:
            managers = self._mo_lookup(
                unit_uuid, "ou/{}/details/manager?inherit_manager=1"
            )
        else:
            managers = self._mo_lookup(unit_uuid, "ou/{}/details/manager")
        # Iterate over all managers, use uuid as key, if more than one
        # distinct uuid shows up in list, rasie an error
        for manager in managers:
            responsibility = {"name": "Intet ansvar"}
            for responsibility in manager["responsibility"]:
                if responsibility["name"] == PRIMARY_RESPONSIBILITY:
                    break
            # TODO: if primary reponsibility is found, this is now selected,
            # otherwise we simply use the last element in the list

            if manager["person"] is not None:
                uuid = manager["person"]["uuid"]
                data = {
                    "Navn": manager["person"]["name"],
                    # 'Ansvar': manager['responsibility'][0]['name'],
                    "Ansvar": responsibility["name"],
                    "uuid": uuid,
                }
                manager_list[uuid] = data
            else:
                # TODO: This is a vacant manager position
                pass
        if len(manager_list) == 0:
            manager = {}
        elif len(manager_list) == 1:
            manager = manager_list[uuid]
        elif len(manager_list) > 1:
            # Currently we do not support multiple managers
            logger.warning("multiple managers not supported for %s", unit_uuid)
            manager = manager_list[uuid]
            # TODO: Fix this...
            # raise Exception('Too many managers')
        return manager

    def read_organisation_people(
        self,
        org_uuid,
        person_type="engagement",
        split_name=True,
        read_all=False,
        skip_past=False,
    ):
        """Read all employees in an ou. If the same employee is listed
        more than once, only the latest listing will be included.
        :param org_uuid: UUID of the OU to find emplyees in.
        :read_all: Read all engagements, not only the present ones.
        :return: The list of emplyoees
        """
        person_list = {}
        if not read_all:
            all_persons = self._mo_lookup(org_uuid, "ou/{}/details/" + person_type)
        else:
            if skip_past:
                validity_times = ["present", "future"]
            else:
                validity_times = ["past", "present", "future"]

            all_persons = []
            for validity in validity_times:
                persons = self._mo_lookup(
                    org_uuid, "ou/{}/details/" + person_type, validity=validity
                )
                all_persons = all_persons + persons

        for person in all_persons:
            uuid = person["person"]["uuid"]
            data = {
                "Ansættelse gyldig fra": person["validity"]["from"],
                "Ansættelse gyldig til": person["validity"]["to"],
                "Person UUID": uuid,
                "Org-enhed": person["org_unit"]["name"],
                "Org-enhed UUID": person["org_unit"]["uuid"],
                "Engagement UUID": person["uuid"],
            }
            if "job_function" in person:
                data["Stillingsbetegnelse"] = person["job_function"]["name"]

            if "engagement_type" in person:
                data["engagement_type_uuid"] = person["engagement_type"]["uuid"]
                data["Engagementstype"] = person["engagement_type"]["name"]

            if "association_type" in person:
                data["Post"] = person["association_type"]["name"]

            # Finally, add name
            if split_name:
                # If a split name i wanted, we prefer to get i directly from MO
                # rather than an algorithmic split.
                if person["person"].get("givenname"):
                    data["Fornavn"] = person["person"].get("givenname")
                    data["Efternavn"] = person["person"].get("surname")
                else:
                    data.update(self._split_name(person["person"]["name"]))
            else:
                data["Navn"] = person["person"]["name"]
            person_list[uuid] = data
        return person_list

    def read_top_units(self, organisation, use_cache=False):
        """Read the ous tha refers directoly to the organisation
        :param organisation: UUID of the organisation
        :return: List of UUIDs of the root OUs in the organisation
        """
        url = "o/{}/children"
        units = self._mo_lookup(organisation, url, use_cache=use_cache)
        return units

    def read_ou_tree(self, org, nodes={}, parent=None):
        """Recursively find all sub-ou's beneath current node
        :param org: The organisation to start the tree from
        :param nodes: Dict with all modes in the tree
        :param parent: The parent of the current node, None if this is root
        :return: A dict with all nodes in tree, top node is named 'root'
        """
        url = "ou/{}/children"
        units = self._mo_lookup(org, url)

        if parent is None:
            nodes["root"] = Node(org)
            parent = nodes["root"]
        for unit in units:
            uuid = unit["uuid"]
            nodes[uuid] = Node(uuid, parent=parent)
            if unit["child_count"] > 0:
                nodes = self.read_ou_tree(uuid, nodes, nodes[uuid])
        return nodes

    def read_ou_root(self, org: str, root: Optional[str] = None):
        """Find all org units below an optional root node
        :param org: The organisation to search in
        :param root: The UUID of an optional root node
        :return: A list of all org units bounded by the root node
        """
        url = f"o/{org}/ou/"
        if root:
            url +=f"?root={root}"
        org_units = self._mo_lookup(None, url)["items"]
        return org_units

    def find_cut_dates(self, uuid, no_past=False):
        """Find dates with changes in engagement history.

        Run through entire history of a user and return a list of dates.

        Args:
            uuid: UUID of the user to find changes for.
            no_past: Do not read engagements in the past.

        Returns:
            list: List of datetimes with changes in engagement history.
        """

        def get_from_date(validity):
            fromdate = datetime.datetime.strptime(validity["from"], "%Y-%m-%d")
            # Python has no love in for dates like 1900
            # Thus we clamp the from date to 1930 if it is before then
            earliest_fromdate = datetime.datetime(1930, 1, 1)
            return max(fromdate, earliest_fromdate)

        def get_to_date(validity):
            if validity["to"]:
                to = datetime.datetime.strptime(validity["to"], "%Y-%m-%d")
                day_after = to + datetime.timedelta(days=1)
                return day_after
            return datetime.datetime(9999, 12, 30, 0, 0)

        mo_engagements = self.read_user_engagement(
            user=uuid, only_primary=True, read_all=True, skip_past=no_past
        )
        validities = map(itemgetter("validity"), mo_engagements)

        dates = set()
        for validity in validities:
            dates.add(get_from_date(validity))
            dates.add(get_to_date(validity))

        date_list = sorted(list(dates))
        logger.debug("List of cut-dates: {}".format(date_list))
        return date_list

    def get_e_username(self, e_uuid, id_it_system):
        for its in self._mo_lookup(e_uuid, "e/{}/details/it"):
            if its["itsystem"]["user_key"] == id_it_system:
                return its["user_key"]
        return ""

    def get_e_addresses(self, e_uuid, scope=None):
        addresses = self._mo_lookup(e_uuid, "e/{}/details/address")
        if scope:
            addresses = filter(
                lambda address: address["address_type"]["scope"] == scope, addresses
            )
        return list(addresses)

    def get_e_address(self, e_uuid, scope):
        return only(self.get_e_addresses(e_uuid, scope), {})

    def get_e_itsystems(self, e_uuid, it_system_uuid=None):
        it_systems = self._mo_lookup(e_uuid, "e/{}/details/it")
        if it_system_uuid:
            it_systems = filter(
                lambda its: its["itsystem"]["uuid"] == it_system_uuid, it_systems
            )
        return list(it_systems)

    def update_user(self, uuid, data):
        payload = {"type": "employee", "uuid": uuid, "data": data}
        return self._mo_post("details/edit", payload)

    def _get_detail(self, uuid, field_type, object_type="e") -> List[Dict]:
        """Get information from /detail for an employee or unit.
        :param uuid: uuid for the object
        :param field_type: detail field type
        :return: dict with the relevant information
        """
        all_data = []
        for validity in ["past", "present", "future"]:
            data = self._mo_lookup(
                uuid,
                object_type + "/{}/details/" + field_type,
                validity=validity,
            )
            all_data += data
        return all_data


if __name__ == "__main__":
    # Example usage: env CLIENT_SECRET=SECRET_UUID_HERE python3 mora_helpers.py
    mh = MoraHelper()
    print("Organisation UUID: " + mh.read_organisation())
    print("Number of users: " + str(len(mh.read_all_users())))



import json
from requests import Session


class IntegrationAbstraction(object):

    def __init__(self, mox_base, system_name, end_marker='STOP'):
        if not mox_base[-1] == '/':
            mox_base = mox_base + '/'
        self.mox_base = mox_base
        self.system_name = system_name
        self.end_marker = end_marker
        self.session = Session()

    def _get_complete_object(self, resource, uuid):
        """ Return a complete LoRa object """
        response = self.session.get(url=self.mox_base + resource + '/' + uuid)
        response.raise_for_status()
        return response.json()

    def _get_attributes(self, resource, uuid):
        """ Return the 'Attributter' part of a LoRa object """
        mox_object = self._get_complete_object(resource, uuid)
        # How to handle multiple 'registreringer'?
        attributes = mox_object[uuid][0]['registreringer'][0]['attributter']
        return attributes

    def _get_integration_data(self, resource, uuid):
        """
        Return the the raw integration data string, no interpretation
        is performed, except for a validation of the data as valid json
        :param resource:
        Path of the service endpoint (str) e.g. /organisation/organisation
        :param uuid:
        uuid of the object to be returned.
        :return: Raw integration data string.
        """
        attributes = self._get_attributes(resource, uuid)
        for key in attributes.keys():
            if key.find('egenskaber') > 0:
                data = attributes[key][0].get('integrationsdata', None)
        if data is not None:
            try:
                json.loads(data)
            except json.decoder.JSONDecodeError:
                raise Exception('Invalid json in integration data')
        return data

    def _set_integration_data(self, resource, uuid, data):
        """
        Updates or creates a raw integration data string in LoRa.
        This is a helper function that will do no attempts of preserving
        existing data. The resulting integration data string is guaranted to
        be valid json.
        :param resource:
        Path of the service endpoint (str) e.g. /organisation/organisation
        :param uuid: uuid of the object to be updated.
        :return: Return none if data is unchanged, otherwise the uuid of the
        object is returned.
        """
        attributter = self._get_attributes(resource, uuid)
        for key in attributter.keys():
            if key.find('egenskaber') > 0:
                current = attributter[key][0].get('integrationsdata', None)
                attributter[key][0]['integrationsdata'] = data
        properties = {'attributter': attributter}

        # Here we should verify that the data is valid json and check
        # whether it is different from the current value
        if current == data:
            return None
        response = self.session.patch(url=self.mox_base + resource +
                                      '/' + uuid, json=properties)
        response.raise_for_status()
        return response.json()

    def read_integration_data(self, resource, uuid):
        """
        Returns the integration data (if any) with the relevant system name.
        :param  resource:
        Path of the service endpoint (str) e.g. /organisation/organisation
        :param uuid: uuid of the object.
        :return: Integration data associated with self.system_name.
        """
        return_value = None

        integration_data = self._get_integration_data(resource, uuid)
        if integration_data is not None:
            structured_data = json.loads(integration_data)
            data = structured_data.get(self.system_name, '')
            end_pos = data.find(self.end_marker)
            if end_pos > -1:
                return_value = json.loads(data[0:end_pos])
        return return_value

    def integration_data_payload(self, resource, value, uuid=None, encode=True):
        """
        Return payload for integration data after update of the relavant system.
        This payload can be included in a bigger payload.
        :param  resource:
        Path of the service endpoint (str) e.g. /organisation/organisation
        :param uuid: uuid of the object.
        :param value: New integration data value.
        """
        if uuid:
            integration_data_string = self._get_integration_data(resource, uuid)
        if uuid and integration_data_string:
            integration_data = json.loads(integration_data_string)
        else:
            integration_data = {}

        value_string = '{}{}'.format(json.dumps(value), self.end_marker)
        integration_data[self.system_name] = value_string
        if encode:
            integration_data_string = json.dumps(integration_data)
        else:
            integration_data_string = integration_data
        return integration_data_string

    def write_integration_data(self, resource, uuid, value):
        """
        Write new integration data for current system.name. If data is already
        present, it will be overwritten.
        :param  resource:
        Path of the service endpoint (str) e.g. /organisation/organisation
        :param uuid: uuid of the object.
        :param value: New integration data value.
        """
        integration_data_string = self._get_integration_data(resource, uuid)
        if integration_data_string is not None:
            integration_data = json.loads(integration_data_string)
        else:
            integration_data = {}

        value_string = '{}{}'.format(json.dumps(value), self.end_marker)

        integration_data[self.system_name] = value_string
        integration_data_string = json.dumps(integration_data)
        self._set_integration_data(resource, uuid, integration_data_string)
        return True

    def find_object(self, resource, key):
        url = self.mox_base + resource + '?integrationsdata=%25{}%25'

        # key_string = repr(key[1:-1]) + self.end_marker
        key_string = json.dumps(key) + self.end_marker
        search_val = json.dumps({self.system_name: key_string})
        search_val = search_val[1:-1]  # Remove { and }
        search_string = search_val.replace('\\', '\\\\')

        response = self.session.get(url=url.format(search_string))
        response.raise_for_status()
        results = response.json()['results'][0]
        if len(results) == 0:
            return_val = None
        elif len(results) == 1:
            return_val = results[0]
        else:
            raise Exception('Inconsistent integration data!')
        return return_val


if __name__ == '__main__':
    ia = IntegrationAbstraction(mox_base='http://localhost:8080',
                                system_name='test',
                                end_marker='Jørgen')

    test_integration_data = json.dumps({"test": "12345Jørgen",
                                        "system": "98Jør\\gen"})

    resource = '/klassifikation/facet'
    uuid = '645e9050-0cad-4138-96b2-6dc89dbdce01'
    print(ia._get_complete_object(resource, uuid))
    print(ia._set_integration_data(resource, uuid, test_integration_data))
    # print(ia.read_integration_data(resource, uuid))
    # print(ia.find_object(resource, '1234'))
    # print(ia.write_integration_data(resource, uuid, '123'))



import asyncio
import json
import sys
from datetime import datetime
from functools import lru_cache, partial
from operator import itemgetter
from typing import Optional, Tuple

import click
from more_itertools import bucket, flatten, unzip
from mox_helpers.mox_helper import create_mox_helper
from mox_helpers.payloads import lora_facet, lora_klasse
from mox_helpers.utils import async_to_sync, dict_map

from ra_utils.load_settings import load_settings


@click.group()
@click.option(
    "--settings-file",
    default="settings.json",
    type=click.Path(resolve_path=True),
    help="Path to the settings.json file",
)
@click.pass_context
def settings_loader(ctx, settings_file: str):
    """MOX Util settings loader.

    This command-level only serves to load settings files for the lower command
    levels, so please refer to the `cli` level for the actual functionality.

    Please run: python mox_util.py cli
    """
    #    # Not available until click 8
    #    settings_file_from_default = (
    #        ctx.get_parameter_source("settings-file") == click.ParameterSource.DEFAULT
    #    )
    settings_file_from_default = True

    settings = {}
    try:
        with click.open_file(settings_file, "rb") as f:
            settings = json.load(f)
    except FileNotFoundError as exp:
        # Only default file is allowed not to exist
        # i.e. if a file argument is passed, the file should exist.
        if not settings_file_from_default:
            raise exp

    def fix_key(key: str, value: str) -> Tuple[str, str]:
        """Fixup keys from settings.json to click format."""
        # NOTE: This is most likely not complete at all
        return key.replace(".", "_"), value

    # Apply fix key to each item tuple in settings
    settings = dict_map(fix_key, settings)

    # Loaded settings are default for cli, thus overriding the coded defaults,
    # but them themselves being overridden by commandline arguments.
    ctx.default_map = {"cli": settings}


@settings_loader.group()
@click.option(
    "--mox-base",
    default="http://localhost:8080",
    show_default=True,
    help="Address of the MOX host",
)
@click.pass_context
def cli(ctx, mox_base: str):
    """MOX Util CLI."""
    ctx.ensure_object(dict)
    # I expect several other keys might be nice to load from settings, or
    # atleast that this multilevel group scheme could be used for exporters.
    ctx.obj["mox.base"] = mox_base


def print_created(uuid: str, created: bool) -> None:
    """Output uuid followed by created or exists.

    The color of the output follows Ansibles changed / unchanged color scheme.
    """
    output_map = {
        False: {
            "message": "exists",
            "color": "green",
        },
        True: {
            "message": "created",
            "color": "yellow",
        },
    }
    output = output_map[created]
    click.secho(uuid + " " + output["message"], fg=output["color"])


def print_changed(uuid: str, changed: bool) -> None:
    """Output uuid followed by not changed or changed.

    The color of the output follows Ansibles changed / unchanged color scheme.
    """
    output_map = {
        False: {
            "message": "unchanged",
            "color": "green",
        },
        True: {
            "message": "changed",
            "color": "yellow",
        },
    }
    output = output_map[changed]
    click.secho(uuid + " " + output["message"], fg=output["color"])


async def ensure_class_exists_helper(
    bvn: str,
    facet_bvn: str,
    title: Optional[str] = None,
    description: Optional[str] = None,
    scope: Optional[str] = None,
    org_uuid: Optional[str] = None,
    org_unit_uuid: Optional[str] = None,
    parent_bvn: Optional[str] = None,
    mox_base: str = "http://localhost:8080",
    dry_run: bool = False,
):
    """Ensure the generated class exists in MOX."""
    mox_helper = await create_mox_helper(mox_base)

    title = title or bvn

    # Fetch default organisation if any, assuming none is set
    org_uuid = org_uuid or await mox_helper.read_element_organisation_organisation(
        bvn="%"
    )
    # Fetch uuids from bvns
    facet_uuid = await mox_helper.read_element_klassifikation_facet(bvn=facet_bvn)
    parent_uuid = None
    if parent_bvn:
        parent_uuid = await mox_helper.read_element_klassifikation_klasse(
            bvn=parent_bvn
        )

    # Generate dict
    klasse = lora_klasse(
        bvn,
        title,
        facet_uuid,
        org_uuid,
        org_unit_uuid=org_unit_uuid,
        description=description,
        scope=scope,
        parent_uuid=parent_uuid,
    )

    # Print for dry run
    if dry_run:
        mox_helper.validate_klassifikation_klasse(klasse)
        message = json.dumps(klasse, indent=4, sort_keys=True)
        click.secho(message, fg="green")
        return

    # POST for non-dry
    response = await mox_helper.get_or_create_klassifikation_klasse(
        klasse, facet=facet_uuid
    )
    return response


@lru_cache(maxsize=None)
@async_to_sync
async def ensure_class_in_lora(facet: str, klasse: str, **kwargs) -> Tuple[str, bool]:
    """Ensures class exists in lora.

    Returns the uuid of the existing class or creates it and returns uuid of the new class.
    Uses mox_utils ensure_class_exists but caches results, so subsequent calls with same parameters will return the correct uuid without any calls to lora.
    Returns a tuple contaning a uuid of the class and a boolean of wether it was created or not.
    Remember that the 'created' boolean is also cached so it will only show if it was created the first time this was called.
    Example:
        uuid, _ = ensure_class_in_lora('org_unit_type', 'Enhed')
        uuid, _ = ensure_class_in_lora('employee_address_type', 'Email', scope = 'EMAIL')
    """
    settings = load_settings()
    mox_base = settings.get("mox.base")
    response = await ensure_class_exists_helper(
        bvn=klasse, facet_bvn=facet, mox_base=mox_base, **kwargs
    )
    return response


@cli.command()
@click.pass_context
@click.option(
    "--bvn",
    "--brugervendt-nøgle",
    required=True,
    help="User key to set on the class.",
)
@click.option("--title", required=True, help="Title to set on the class.")
@click.option(
    "--facet-bvn", required=True, help="User key of the facet to bind the class to."
)
@click.option("--description", help="Description to set on the class.")
@click.option("--scope", help="Scope for the class.")
@click.option("--org-uuid", "--organisation", help="Organisation to bind the class to.")
@click.option(
    "--org-unit-uuid", "--organisation-unit", help="Organisation Unit to own the class."
)
@click.option("--parent-bvn", help="User key of another class to put this under.")
@click.option(
    "--dry-run",
    default=False,
    is_flag=True,
    help="Dry run and print the generated object.",
)
@async_to_sync
async def ensure_class_exists(
    ctx,
    bvn: str,
    title: str,
    facet_bvn: str,
    description: str,
    scope: str,
    org_uuid: str,
    org_unit_uuid: str,
    parent_bvn: str,
    dry_run: bool,
):
    uuid, created = await ensure_class_exists_helper(
        bvn=bvn,
        title=title,
        facet_bvn=facet_bvn,
        description=description,
        scope=scope,
        org_uuid=org_uuid,
        org_unit_uuid=org_unit_uuid,
        parent_bvn=parent_bvn,
        mox_base=ctx.obj["mox.base"],
        dry_run=dry_run,
    )
    print_created(uuid, created)


async def ensure_class_value_helper(
    variable: str,
    new_value: str,
    mox_base: str = "localhost:8080",
    bvn: str = None,
    uuid: str = None,
    dry_run: bool = False,
):
    """Ensures a value of a class is as expected. """
    mox_helper = await create_mox_helper(mox_base)

    if bvn:
        try:
            uuid = await mox_helper.read_element_klassifikation_klasse({"bvn": bvn})
        except:
            message = "No class with bvn={} was found.".format(bvn)
            click.secho(message, fg="red")
            return
    if uuid is None:
        raise click.ClickException("Must provide either bvn or UUID")

    klasse = await mox_helper.search_klassifikation_klasse({"UUID": uuid})
    klasse = klasse[0]["registreringer"][0]
    klasse = {
        item: klasse.get(item) for item in ("attributter", "relationer", "tilstande")
    }

    virkning = {"from": datetime.now().strftime("%Y-%m-%d"), "to": "infinity"}

    def check_value(variable, new_value, o):
        """Recurse through object to ensure correct value "new_value" in "variable"."""
        seeded_check_value = partial(check_value, variable, new_value)
        if isinstance(o, dict):
            if variable in o:
                if o[variable] == new_value:
                    return o, False
                o[variable] = new_value
                o["virkning"] = virkning
                return o, True
            keys, values = unzip(o.items())
            values, changed = unzip(map(seeded_check_value, values))
            return dict(zip(keys, values)), any(changed)
        elif isinstance(o, list):
            values, changed = unzip(map(seeded_check_value, o))
            return list(values), any(changed)
        elif isinstance(o, tuple):
            values, changed = unzip(map(seeded_check_value, o))
            return tuple(values), any(changed)
        else:
            return o, False

    if variable == "ejer":
        owner = klasse.get("relationer").get("ejer")
        changed = False
        if not owner:
            changed = True
        else:
            try:
                old_owner = owner[0].get("uuid")
                changed = old_owner != new_value
            except IndexError:
                changed = True

        if changed:
            klasse["relationer"]["ejer"] = [
                {
                    "uuid": new_value,
                    "virkning": virkning,
                    "objekttype": "OrganisationEnhed",
                }
            ]
    else:
        klasse, changed = check_value(variable, new_value, klasse)
    # Print for dry run
    if dry_run:
        mox_helper.validate_klassifikation_klasse(klasse)
        message = json.dumps(klasse, indent=4, sort_keys=True)
        click.secho(message, fg="green")
        return

    # POST for non-dry
    if changed:
        uuid = await mox_helper.update_klassifikation_klasse(uuid, klasse)
    print_changed(uuid, changed)


@cli.command()
@click.pass_context
@click.option(
    "--bvn",
    "--brugervendt-nøgle",
    help="User key for the class.",
)
@click.option(
    "--uuid",
    help="UUID for the class.",
)
@click.option("--variable", required=True, help="variable to be checked/updated")
@click.option(
    "--new_value", required=True, help="Value which should be checked/updated."
)
@click.option(
    "--dry-run",
    default=False,
    is_flag=True,
    help="Dry run and print the generated object.",
)
@async_to_sync
async def ensure_class_value(
    ctx,
    bvn: str,
    uuid: str,
    variable: str,
    new_value: str,
    dry_run: bool,
):
    await ensure_class_value_helper(
        mox_base=ctx.obj["mox.base"],
        bvn=bvn,
        uuid=uuid,
        variable=variable,
        new_value=new_value,
        dry_run=dry_run,
    )


@cli.command()
@click.pass_context
@click.option(
    "--bvn",
    "--brugervendt-nøgle",
    required=True,
    help="User key to set on the facet.",
)
@click.option("--description", help="Description to set on the facet.")
@click.option("--org-uuid", "--organisation", help="Organisation to bind the facet to.")
@click.option(
    "--dry-run",
    default=False,
    is_flag=True,
    help="Dry run and print the generated object.",
)
@async_to_sync
async def ensure_facet_exists(
    ctx, bvn: str, description: str, org_uuid: str, dry_run: bool
):
    """Ensure the generated facet exists in MOX."""
    mox_helper = await create_mox_helper(ctx.obj["mox.base"])

    # Fetch default organisation if any, assuming none is set
    org_uuid = org_uuid or await mox_helper.read_element_organisation_organisation(
        bvn="%"
    )

    # Generate dict
    facet = lora_facet(bvn, org_uuid, description)

    # Print for dry run
    if dry_run:
        mox_helper.validate_klassifikation_facet(facet)
        print(json.dumps(facet, indent=4, sort_keys=True))
        return

    # POST for non-dry
    uuid, created = await mox_helper.get_or_create_klassifikation_facet(facet)
    print_created(uuid, created)


@cli.command()
@click.pass_context
@click.option(
    "--dry-run",
    default=False,
    is_flag=True,
    help="Dry run and print the generated object.",
)
@click.argument("filename", required=True, type=click.Path(exists=True))
@async_to_sync
async def bulk_ensure(
    ctx,
    dry_run: bool,
    filename: str,
):
    """Ensure the entries in the json file exists in MOX.

    Currently only bulk loads classes
    """
    mox_helper = await create_mox_helper(ctx.obj["mox.base"])

    # Load file and fetch
    with open(filename) as json_file:
        data = json.load(json_file)

    # Construct classes by applies __apply_to_all__ to all elements within a
    # single block of classes and flattening the structure to be a simple
    # list of classes
    def construct_entry(bvn, item, apply_to_all):
        return {**item, **apply_to_all, "bvn": bvn}

    def construct_block(block):
        apply_to_all = block.pop("__apply_to_all__", {})
        classes = map(
            lambda entry: construct_entry(*entry, apply_to_all), block.items()
        )
        return classes

    facets = []
    if "facets" in data:
        facets = flatten(map(construct_block, data["facets"]))
        facets = list(facets)

    classes = []
    if "classes" in data:
        classes = flatten(map(construct_block, data["classes"]))
        classes = list(classes)

    # Fetch default organisation
    org_uuid = None
    org_uuid = org_uuid or await mox_helper.read_element_organisation_organisation(
        bvn="%"
    )

    def enrich_with_org_unit(entry):
        entry["org_uuid"] = org_uuid
        return entry

    # Enrich facets with default organisation
    facets = map(enrich_with_org_unit, facets)

    # Translate facet json to lora_facet
    facets = map(lambda facet: lora_facet(**facet), facets)

    # Prepare to output
    facets = list(facets)

    # Print for dry run
    if dry_run:
        for facet in facets:
            mox_helper.validate_klassifikation_facet(facet)
            message = json.dumps(facet, indent=4, sort_keys=True)
            click.secho(message, fg="green")
        return

    # POST for non-dry
    tasks = list(map(mox_helper.get_or_create_klassifikation_facet, facets))
    results = await asyncio.gather(*tasks)
    for uuid, created in results:
        print_created(uuid, created)

    # Find all unique facet bvns used by the classes, and translate to UUIDs
    required_facets = set(map(itemgetter("facet"), classes))

    async def construct_facet_bvn_to_uuid_map(facet_bvns):
        async def create_bvn_to_uuid_tuple(facet_bvn):
            return (
                facet_bvn,
                await mox_helper.read_element_klassifikation_facet(bvn=facet_bvn),
            )

        tasks = list(map(create_bvn_to_uuid_tuple, facet_bvns))
        return dict(await asyncio.gather(*tasks))

    facet_map = await construct_facet_bvn_to_uuid_map(required_facets)

    async def enrich_classes(classes):
        # Find all unique parent bvns used by the classes, and translate to UUIDs
        required_parents = set(
            {clazz["parent"] for clazz in classes if "parent" in clazz}
        )

        async def construct_parent_bvn_to_uuid_map(parent_bvns):
            async def create_bvn_to_uuid_tuple(parent_bvn):
                return (
                    parent_bvn,
                    await mox_helper.read_element_klassifikation_klasse(bvn=parent_bvn),
                )

            tasks = list(map(create_bvn_to_uuid_tuple, parent_bvns))
            return dict(await asyncio.gather(*tasks))

        parent_map = await construct_parent_bvn_to_uuid_map(required_parents)

        # Enrich classes with default organisation
        classes = map(enrich_with_org_unit, classes)

        # Translate class facet to facet_uuid
        def class_facet_to_facet_uuid(clazz):
            facet_bvn = clazz.pop("facet")
            clazz["facet_uuid"] = facet_map[facet_bvn]
            return clazz

        classes = map(class_facet_to_facet_uuid, classes)

        # Translate class parent to parent_uuid
        def class_parent_to_parent_uuid(clazz):
            parent_bvn = clazz.pop("parent", None)
            if parent_bvn:
                clazz["parent_uuid"] = parent_map[parent_bvn]
            return clazz

        classes = map(class_parent_to_parent_uuid, classes)

        return classes

    # Partition into buckets by layer
    def set_layer(clazz):
        if "__layer__" not in clazz:
            clazz["__layer__"] = 1
        return clazz

    classes = map(set_layer, classes)
    buckets = bucket(classes, key=itemgetter("__layer__"))
    layers = sorted(list(buckets))
    for layer in layers:
        classes = list(buckets[layer])
        classes = await enrich_classes(classes)

        # Remove the layer key
        def remove_key(key):
            def worker(clazz):
                del clazz[key]
                return clazz

            return worker

        classes = map(remove_key("__layer__"), classes)

        # Translate class json to lora_klasse
        classes = map(lambda clazz: lora_klasse(**clazz), classes)

        # Prepare to output
        classes = list(classes)

        # Print for dry run
        if dry_run:
            for clazz in classes:
                mox_helper.validate_klassifikation_klasse(clazz)
                message = json.dumps(clazz, indent=4, sort_keys=True)
                click.secho(message, fg="green")
            return

        # POST for non-dry
        tasks = list(map(mox_helper.get_or_create_klassifikation_klasse, classes))
        results = await asyncio.gather(*tasks)
        for uuid, created in results:
            print_created(uuid, created)


@cli.command()
@click.option(
    "--exit-print-on-success",
    show_default=True,
    type=click.STRING,
    help="String to print if able to connect",
)
@click.option(
    "--exit-print-on-error",
    show_default=True,
    default="Unable to connect",
    type=click.STRING,
    help="String to print if unable to connect",
)
@click.option(
    "--exit-code-on-success",
    show_default=True,
    default=0,
    type=click.IntRange(0, 255),
    help="Exit code if able to connect",
)
@click.option(
    "--exit-code-on-error",
    show_default=True,
    default=1,
    type=click.IntRange(0, 255),
    help="Exit code if unable to connect",
)
@click.pass_context
def check_connection(
    ctx,
    exit_print_on_success: str,
    exit_print_on_error: str,
    exit_code_on_success: int,
    exit_code_on_error: int,
):
    """Check whether a connection can be established to mox."""
    output_map = {
        False: {
            "exit_code": exit_code_on_error,
            "message": exit_print_on_error,
            "color": "red",
        },
        True: {
            "exit_code": exit_code_on_success,
            "message": exit_print_on_success,
            "color": "green",
        },
    }

    @async_to_sync
    async def is_up() -> bool:
        mox_helper = await create_mox_helper(
            ctx.obj["mox.base"], generate_methods=False
        )
        return await mox_helper.check_connection()

    output = output_map[is_up()]
    if output["message"]:
        click.secho(output["message"], fg=output["color"])
    sys.exit(output["exit_code"])


@cli.command()
@click.pass_context
def noop(ctx):
    """Do nothing."""
    pass


if __name__ == "__main__":
    settings_loader(auto_envvar_prefix="MOXUTIL")



#
# Copyright (c) Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#

import logging
from uuid import uuid4, UUID
from urllib.parse import urljoin
from requests import Session, HTTPError
from datetime import datetime, timedelta

from integration_abstraction.integration_abstraction import IntegrationAbstraction

from os2mo_data_import.mora_data_types import (
    OrganisationUnitType,
    TerminationType,
    EngagementTerminationType,
    EmployeeType
)

from os2mo_data_import.mox_data_types import (
    Organisation,
    Klassifikation,
    Facet,
    Klasse,
    Itsystem
)
from os2mo_helpers.mora_helpers import MoraHelper

logger = logging.getLogger("moImporterUtilities")


class ImportUtility(object):
    """
    ImportUtility
    TODO: This class relies heavily on asserts, which are gennerally not wanted in
    this context. We should create meaningfull exceptions and raise these when
    needed.
    """

    def __init__(self, system_name, end_marker, mox_base, mora_base,
                 demand_consistent_uuids, store_integration_data=False,
                 dry_run=False):

        # Import Params
        self.demand_consistent_uuids = demand_consistent_uuids
        self.store_integration_data = store_integration_data
        if store_integration_data:
            self.ia = IntegrationAbstraction(mox_base, system_name, end_marker)

        # Service endpoint base
        self.mox_base = mox_base
        self.mora_base = mora_base

        # Session
        self.mh = MoraHelper(self.mora_base, use_cache=False)
        self.session = Session()

        # Placeholder for UUID import
        self.organisation_uuid = None

        # Existing UUIDS
        # TODO: More elegant version of this please
        self.existing_uuids = []

        # UUID map
        self.inserted_organisation = {}
        self.inserted_facet_map = {}
        self.inserted_klasse_map = {}
        self.inserted_itsystem_map = {}
        self.inserted_org_unit_map = {}
        self.inserted_employee_map = {}

        # Deprecated
        self.dry_run = dry_run

    def import_organisation(self, reference, organisation):
        """
        Convert organisation to OIO formatted post data
        and import into the MOX datastore.

        :param str reference: Reference to the user defined identifier
        :param object organisation: Organisation object

        :returns: Inserted UUID
        :rtype: str/uuid
        """

        if not isinstance(organisation, Organisation):
            raise TypeError("Not of type Organisation")

        resource = "organisation/organisation"

        integration_data = self._integration_data(
            resource=resource,
            reference=reference,
            payload={}
        )

        organisation.integration_data = integration_data

        payload = organisation.build()

        if organisation.uuid is not None:
            organisation_uuid = organisation.uuid
        else:
            organisation_uuid = integration_data.get('uuid', None)

        self.organisation_uuid = self.insert_mox_data(
            resource=resource,
            data=payload,
            uuid=organisation_uuid
        )

        # Global validity
        self.date_from = organisation.date_from
        self.date_to = organisation.date_to

        return self.organisation_uuid

    def import_klassifikation(self, reference, klassifikation):
        """
        Begin import of klassifikation

        :param str reference: Reference to the user defined identifier
        :param object klassifikation: Klassifikation object

        :returns: Inserted UUID
        :rtype: str/uuid
        """
        if not isinstance(klassifikation, Klassifikation):
            raise TypeError("Not of type Klassifikation")

        resource = "klassifikation/klassifikation"

        klassifikation.organisation_uuid = self.organisation_uuid

        integration_data = self._integration_data(
            resource=resource,
            reference=reference,
            payload={}
        )

        klassifikation.integration_data = integration_data
        payload = klassifikation.build()

        klassifikation_uuid = integration_data.get('uuid', None)

        self.klassifikation_uuid = self.insert_mox_data(
            resource=resource,
            data=payload,
            uuid=klassifikation_uuid
        )

        return self.klassifikation_uuid

    def import_facet(self, reference, facet):
        """
        Begin import of facet

        :param str reference: Reference to the user defined identifier
        :param object facet: Facet object

        :returns: Inserted UUID
        :rtype: str/uuid
        """

        if not isinstance(facet, Facet):
            raise TypeError("Not of type Facet")

        resource = "klassifikation/facet"

        facet.organisation_uuid = self.organisation_uuid
        facet.klassifikation_uuid = self.klassifikation_uuid

        # NEED TO BE FIXED
        facet.date_from = self.date_from
        facet.date_to = self.date_to

        integration_data = self._integration_data(
            resource=resource,
            reference=reference,
            payload={}
        )

        facet.integration_data = integration_data
        payload = facet.build()

        facet_uuid = integration_data.get('uuid', None)

        self.inserted_facet_map[reference] = self.insert_mox_data(
            resource=resource,
            data=payload,
            uuid=facet_uuid
        )

        return self.inserted_facet_map[reference]

    def import_klasse(self, reference, klasse):
        """
        Insert a klasse object

        Begin import of klassifikation

        :param str reference: Reference to the user defined identifier
        :param object organisation: Organisation object

        :returns: Inserted UUID
        :rtype: str/uuid
        """

        if not isinstance(klasse, Klasse):
            raise TypeError("Not of type Facet")

        uuid = klasse.uuid
        facet_ref = klasse.facet_type_ref

        facet_uuid = self.inserted_facet_map.get(facet_ref)

        if not facet_uuid:
            error_message = "Facet ref: {} does not exist for {}".format(
                facet_ref, klasse
            )
            logger.error(error_message)
            raise KeyError(error_message)

        resource = "klassifikation/klasse"

        klasse.organisation_uuid = self.organisation_uuid
        klasse.facet_uuid = facet_uuid
        klasse.date_from = self.date_from
        klasse.date_to = self.date_to

        integration_data = self._integration_data(
            resource=resource,
            reference=reference,
            payload={}
        )

        if 'uuid' in integration_data:
            klasse_uuid = integration_data['uuid']
            assert(uuid is None or klasse_uuid == uuid)
        else:
            if uuid is None:
                klasse_uuid = None
            else:
                klasse_uuid = uuid

        klasse.integration_data = integration_data
        payload = klasse.build()

        import_uuid = self.insert_mox_data(
            resource="klassifikation/klasse",
            data=payload,
            uuid=klasse_uuid
        )
        assert(uuid is None or import_uuid == str(klasse_uuid))
        self.inserted_klasse_map[reference] = import_uuid

        return self.inserted_klasse_map[reference]

    def import_itsystem(self, reference, itsystem):
        """
        Create IT System

        :param str reference: Reference to the user defined identifier
        :param object organisation: Organisation object

        :returns: Inserted UUID
        :rtype: str/uuid
        """

        if not isinstance(itsystem, Itsystem):
            raise TypeError("Not of type Itsystem")

        resource = 'organisation/itsystem'

        itsystem.organisation_uuid = self.organisation_uuid
        itsystem.date_from = self.date_from
        itsystem.date_to = self.date_to

        integration_data = self._integration_data(
            resource=resource,
            reference=reference,
            payload={}
        )

        if 'uuid' in integration_data:
            itsystem_uuid = integration_data['uuid']
        else:
            itsystem_uuid = None

        itsystem.integration_data = integration_data
        payload = itsystem.build()

        self.inserted_itsystem_map[reference] = self.insert_mox_data(
            resource=resource,
            data=payload,
            uuid=itsystem_uuid
        )

        return self.inserted_itsystem_map[reference]

    def import_org_unit(self, reference, organisation_unit, details=[]):
        """
        Insert organisation unit and details

        .. note::
            Optional data objects are relational objects which
            belong to the organisation unit, such as an address type

        :param str reference: Reference to the user defined identifier
        :param object organisation_unit: Organisation object
        :param list details: List of details
        :returns: Inserted UUID
        :rtype: str/uuid
        """

        if not isinstance(organisation_unit, OrganisationUnitType):
            raise TypeError("Not of type OrganisationUnitType")

        if reference in self.inserted_org_unit_map:
            return False

        resource = 'organisation/organisationenhed'

        # payload = self.build_mo_payload(organisation_unit_data)
        parent_ref = organisation_unit.parent_ref
        if parent_ref:
            parent_uuid = self.inserted_org_unit_map.get(parent_ref)
            organisation_unit.parent_uuid = parent_uuid

        if not organisation_unit.parent_uuid:
            organisation_unit.parent_uuid = self.organisation_uuid

        type_ref_uuid = self.inserted_klasse_map.get(
            organisation_unit.type_ref
        )

        if hasattr(organisation_unit, "time_planning_ref"):
            organisation_unit.time_planning_ref_uuid = self.inserted_klasse_map.get(
                organisation_unit.time_planning_ref
            )

        if hasattr(organisation_unit, 'org_unit_level_ref'):
            organisation_unit.org_unit_level_uuid = self.inserted_klasse_map.get(
                organisation_unit.org_unit_level_ref
            )

        organisation_unit.type_ref_uuid = type_ref_uuid

        payload = organisation_unit.build()
        payload = self._integration_data(
            resource=resource,
            reference=reference,
            payload=payload,
            encode_integration=False
        )

        if 'uuid' in payload:
            if payload['uuid'] in self.existing_uuids:
                logger.info('Re-import org-unit: {}'.format(payload['uuid']))
                re_import = 'NO'
                resource = 'details/edit'
                payload_keys = list(payload.keys())
                payload['data'] = {}
                for key in payload_keys:
                    payload['data'][key] = payload[key]
                    del payload[key]
                payload['type'] = 'org_unit'
            else:
                re_import = 'NEW'
                logger.info('New unit - Forced uuid: {}'.format(payload['uuid']))
                resource = 'ou/create'
        else:
            re_import = 'NEW'
            logger.info('New unit, random uuid')
            resource = 'ou/create'

        logger.debug('Unit payload: {}'.format(payload))

        uuid = self.insert_mora_data(
            resource=resource,
            data=payload
        )

        if 'uuid' in payload:
            assert (uuid == payload['uuid'])
        if not uuid:
            raise ConnectionError("Something went wrong")

        # Add to the inserted map
        self.inserted_org_unit_map[reference] = uuid

        data = {}
        data['address'] = self.mh._get_detail(uuid, 'address', object_type='ou')

        # Build details (if any)
        details_payload = []

        for detail in details:
            detail.org_unit_uuid = uuid

            date_from = detail.date_from

            if not date_from:
                date_from = organisation_unit.date_from

            build_detail = self.build_detail(
                detail=detail
            )

            if not build_detail:
                continue

            found_hit = self._payload_compare(build_detail, data)
            logger.debug('Found hit for {}: {}'.format(uuid, found_hit))
            if not found_hit and re_import == 'NO':
                re_import = 'YES'

            # TODO: SHOLD WE UPDATE FROM TO TODAY IN CASE OF RE-IMPORT?
            if re_import == 'YES':
                valid_from = datetime.now().strftime('%Y-%m-%d')  # today
                build_detail['validity']['from'] = valid_from
            details_payload.append(build_detail)

        if re_import == 'YES':
            logger.info('Terminating details for unit {}'.format(uuid))
            for item in data['address']:
                self._terminate_details(item['uuid'], 'address')
        if re_import in ('YES', 'NEW'):
            logger.info('Re-import unit: {}'.format(re_import))
            self.insert_mora_data(
                resource="details/create",
                data=details_payload
            )

        return uuid

    def import_employee(self, reference, employee, details=[]):
        """
        Import employee

        :param str reference: Reference to the user defined identifier
        :param object employee: Employee object
        :param list details: List of details
        :returns: Inserted UUID
        :rtype: str/uuid
        """
        if not isinstance(employee, EmployeeType):
            raise TypeError("Not of type EmployeeType")

        employee.org_uuid = self.organisation_uuid
        payload = employee.build()
        mox_resource = 'organisation/bruger'

        integration_data = self._integration_data(
            resource=mox_resource,
            reference=reference,
            payload=payload,
            encode_integration=False
        )

        if 'uuid' in payload and payload['uuid'] in self.existing_uuids:
            logger.info('Re-import employee {}'.format(payload['uuid']))
            re_import = 'NO'
        else:
            re_import = 'NEW'
            logger.info('New employee, uuid {}'.format(payload.get('uuid')))

        # We unconditionally create or update the user, this should
        # ensure that we are always updated with correct current information.
        mora_resource = "e/create"
        uuid = self.insert_mora_data(
            resource=mora_resource,
            data=integration_data
        )

        if 'uuid' in integration_data:
            assert (uuid == integration_data['uuid'])

        # Add uuid to the inserted employee map
        self.inserted_employee_map[reference] = uuid

        data = {}
        data['it'] = self.mh._get_detail(uuid, 'it')
        data['role'] = self.mh._get_detail(uuid, 'role')
        data['leave'] = self.mh._get_detail(uuid, 'leave')
        data['address'] = self.mh._get_detail(uuid, 'address')
        data['manager'] = self.mh._get_detail(uuid, 'manager')
        data['engagement'] = self.mh._get_detail(uuid, 'engagement')
        data['association'] = self.mh._get_detail(uuid, 'association')

        # In case of en explicit termination, we terminate the employee or
        # employment and return imidiately.
        for detail in details:
            if isinstance(detail, TerminationType):
                logger.info(
                    'Explicit termination of eng {} from {}'.format(
                        uuid, detail.date_from
                    )
                )
                self._terminate_employee(uuid, date_from=detail.date_from)
                return uuid
            if isinstance(detail, EngagementTerminationType):
                logger.info(
                    'Explicit termination of eng {}'.format(
                        detail.engagement_uuid
                    )
                )
                self._terminate_details(detail.engagement_uuid, 'engagement')
                return uuid

        if details:
            additional_payload = []
            for detail in details:

                if not detail.date_from:
                    detail.date_from = self.date_from

                # Create payload (as dict)
                detail_payload = self.build_detail(
                    detail=detail,
                    employee_uuid=uuid
                )

                if not detail_payload:
                    continue

                # If we do not have existing data, the new data should be imported
                if len(data[detail_payload['type']]) == 0 and re_import == 'NO':
                    re_import = 'UPDATE'
                elif data[detail_payload['type']]:
                    found_hit = self._payload_compare(detail_payload, data)
                    if not found_hit:
                        re_import = 'YES'
                additional_payload.append(detail_payload)

            for item in additional_payload:
                valid_from = item['validity']['from']
                valid_to = item['validity']['to']
                now = datetime.now()
                py_from = datetime.strptime(valid_from, '%Y-%m-%d')
                if valid_to is not None:
                    py_to = datetime.strptime(valid_to, '%Y-%m-%d')
                else:
                    py_to = datetime.strptime('2200-01-01', '%Y-%m-%d')

                logger.debug(
                    'Py-from:{}, Py-to:{}, Now:{}'.format(py_from, py_to, now)
                )
                if re_import == 'YES' and py_from < now and py_to > now:
                    logger.debug('Updating valid_from')
                    valid_from = datetime.now().strftime('%Y-%m-%d')  # today
                    item['validity']['from'] = valid_from

            logger.info('Re-import: {}'.format(re_import))

            if re_import == 'YES':
                logger.info('Non-explicit termination: {}'.format(uuid))
                self._terminate_employee(uuid)

            if re_import in ('YES', 'NEW', 'UPDATE'):
                self.insert_mora_data(
                    resource="details/create",
                    data=additional_payload
                )

        return uuid

    def build_detail(self, detail, employee_uuid=None):
        """
        Build detail payload

        :param MoType detail: Detail object

        .. note::
            A detail can be one of the following types:

                - address
                - asso
                - role
                - itsystem
                - engagement
                - manager

            :Reference: :mod:`os2mo_data_import.mora_data_types`

        :param str employee_uuid: (Option) Employee uuid if it exists

        :return: Detail POST data payload
        :rtype: dict
        """

        if employee_uuid:
            detail.person_uuid = employee_uuid

        common_attributes = [
            ("type_ref", "type_ref_uuid"),
            ("job_function_ref", "job_function_uuid"),
            ("address_type_ref", "address_type_uuid"),
            ("manager_level_ref", "manager_level_uuid")
        ]

        for check_value, set_value in common_attributes:
            if not hasattr(detail, check_value):
                continue

            uuid = self.inserted_klasse_map.get(
                getattr(detail, check_value)
            )

            if not uuid:
                if self.store_integration_data:
                    klasse_res = 'klassifikation/klasse'
                    uuid = self.ia.find_object(klasse_res, getattr(detail, check_value))
                else:
                    # print('Detail: {}, check_value: {}'.format(detail, check_value))
                    pass

            setattr(detail, set_value, uuid)

        # Uncommon attributes
        if hasattr(detail, "visibility_ref"):
            detail.visibility_ref_uuid = self.inserted_klasse_map.get(
                detail.visibility_ref
            )

        if hasattr(detail, 'primary_ref'):
            detail.primary_uuid = self.inserted_klasse_map.get(
                detail.primary_ref
            )

        if hasattr(detail, "org_unit_ref"):
            detail.org_unit_uuid = self.inserted_org_unit_map.get(
                detail.org_unit_ref
            )

        if hasattr(detail, "organisation_uuid"):
            detail.organisation_uuid = self.organisation_uuid

        if hasattr(detail, "itsystem_ref"):
            detail.itsystem_uuid = self.inserted_itsystem_map.get(
                detail.itsystem_ref
            )

        if hasattr(detail, "responsibilities"):
            detail.responsibilities = [
                self.inserted_klasse_map[reference]
                for reference in detail.responsibility_list
            ]

        return detail.build()

    def _integration_data(self, resource, reference, payload={},
                          encode_integration=True):
        """
        Update the payload with integration data. Checks if an object with this
        integration data already exists. In this case the uuid of the exisiting
        object is put into the payload. If a supplied uuid is inconsistent with the
        uuid found from integration data, an exception is raised.

        :param resource:
        LoRa resource URL.

        :param referece:
        Unique label that will be stored in the integration data to identify the
        object on re-import.

        :param payload:
        The supplied payload will be updated with values for integration and uuid (if
        the integration data was found from an earlier import). For MO objects,
        payload will typically be pre-populated and will then be ready for import
        when returned. For MOX objects, the initial payload  will typically be empty,
        and the returned values can be fed to the relevant adapter.

        :param encode_integration:
        If True, the integration data will be returned in json-encoded form.

        :return:
        The original payload updated with integration data and object uuid, if the
        object was already imported.
        """
        # TODO: We need to have a list of all objects with integration data to
        # be able to make a list of objects that has disappeared

        if self.store_integration_data:
            uuid = self.ia.find_object(resource, reference)
            if uuid:
                if 'uuid' in payload:
                    if self.demand_consistent_uuids:
                        assert(payload['uuid'] == uuid)
                    else:
                        payload['uuid'] = uuid
                payload['uuid'] = uuid
                self.existing_uuids.append(uuid)

            payload['integration_data'] = self.ia.integration_data_payload(
                resource,
                reference,
                uuid,
                encode_integration
            )
        return payload

    def insert_mox_data(self, resource, data, uuid=None):

        service_url = urljoin(
            base=self.mox_base,
            url=resource
        )

        if uuid:
            update_url = "{service}/{uuid}".format(
                service=service_url,
                uuid=uuid
            )

            response = self.session.put(
                url=update_url,
                json=data
            )

            if response.status_code != 200:
                logger.error(
                    'Mox put. Response: {}, data'.format(response.text, data)
                )
                raise HTTPError("Inserting mox data failed")

        else:
            response = self.session.post(
                url=service_url,
                json=data
            )

            if response.status_code != 201:
                logger.error(
                    'Mox post. Response: {}, data'.format(response.text, data)
                )

                raise HTTPError("Inserting mox data failed")

        response_data = response.json()
        return response_data["uuid"]

    def insert_mora_data(self, resource, data, uuid=None):

        # TESTING
        if self.dry_run:
            uuid = uuid4()
            return str(uuid)

        response = self.mh._mo_post(
            url=resource,
            payload=data,
            force=True,
        )

        if response.status_code == 400:
            error = response.json()['description']
            if error.find('does not give raise to a new registration') > 0:
                uuid_start = error.find('with id [')
                uuid = error[uuid_start+9:uuid_start+45]
                try:
                    UUID(uuid, version=4)
                except ValueError:
                    raise Exception('Unable to read uuid')
            else:
                logger.error(
                    'MO post. Error: {}, data: {}'.format(error, data)
                )
                raise HTTPError("Inserting mora data failed")
        elif response.status_code not in (200, 201):
            logger.error(
                'MO post. Response: {}, data: {}'.format(response.text, data)
            )
            raise HTTPError("Inserting mora data failed")
        else:
            uuid = response.json()
        return uuid

    def _terminate_employee(self, uuid, date_from=None):
        endpoint = 'e/{}/terminate'
        yesterday = datetime.now() - timedelta(days=1)
        if date_from:
            to = date_from
        else:
            to = yesterday.strftime('%Y-%m-%d')

        payload = {
            'terminate_all': True,
            'validity': {
                'to': to
            }
        }
        resource = endpoint.format(uuid)

        self.insert_mora_data(
            resource=resource,
            data=payload
        )
        return uuid

    def _terminate_details(self, uuid, detail_type):
        logger.info('Terminate detail {}:  {}'.format(uuid, detail_type))
        yesterday = datetime.now() - timedelta(days=1)
        payload = {
            'type': detail_type,
            'uuid': uuid,
            'validity': {
                'to': yesterday.strftime('%Y-%m-%d')
            }
        }
        logger.debug('Terminate detail payload: {}'.format(payload))
        try:
            self.insert_mora_data(
                resource='details/terminate',
                data=payload
            )
        except HTTPError:
            logger.info('Tried to terminate a non-existing uuid')
        return uuid

    def _std_compare(self, item_payload, data_item, extra_field=None):
        """ Helper for _payload_compare, performs the checks that are identical
        for most object types.
        :param item_payload: The new payload data.
        :param data_item: The existing set of data.
        :param extra_field: If not None the comparison will also be done on this
        field, otherwise the comparison is only performed on uuid and validity.
        :return: True if identical, otherwise False
        """
        identical = (
            (data_item['org_unit']['uuid'] == item_payload['org_unit']['uuid']) and
            (data_item['validity']['from'] == item_payload['validity']['from']) and
            (data_item['validity']['to'] == item_payload['validity']['to'])
        )
        if extra_field is not None:
            identical = (
                identical and
                data_item[extra_field]['uuid'] == item_payload[extra_field]['uuid']
            )
        return identical

    def _payload_compare(self, item_payload, data):
        """ Compare an exising data-set with a new payload and tell whether
        the new payload is different from the exiting data.
        :param item_payload: New the payload data.
        :param data_item: The existing set of data.
        :param extra_field: If not None the comparison will also be done on this
        field, otherwise the comparison is only performed on uuid and validity.
        :return: True if identical, otherwise False
        """
        data_type = item_payload['type']

        logger.debug(
            'Payload compare. item_payload: {}, data: {}'.format(item_payload, data)
        )

        found_hit = False
        if data_type == 'engagement':
            # In priciple, we should be able to re-calculate the hash stored in
            # integration data and compare directly from that.
            for data_item in data[data_type]:
                if (self._std_compare(item_payload, data_item, 'job_function') and
                    item_payload['user_key'] == data_item['user_key'] and
                    item_payload['fraction'] == data_item['fraction']):
                    found_hit = True

        elif data_type == 'role':
            for data_item in data[data_type]:
                if self._std_compare(item_payload, data_item, 'role_type'):
                    found_hit = True
                    logger.debug('Found hit in role')
        elif data_type == 'leave':
            for data_item in data[data_type]:
                if ((data_item['validity']['from'] ==
                     item_payload['validity']['from']) and

                    (data_item['validity']['to'] ==
                     item_payload['validity']['to'])):
                    found_hit = True
                    logger.debug('Found hit in leave')
        elif data_type == 'it':
            for data_item in data[data_type]:
                if (
                    (data_item['validity']['from'] ==
                     item_payload['validity']['from']) and

                    (data_item['validity']['to'] ==
                     item_payload['validity']['to']) and

                    (data_item['itsystem']['uuid'] ==
                     item_payload['itsystem']['uuid'])
                ):
                    found_hit = True
                    logger.debug('Found hit in it')
        elif data_type == 'address':
            for data_item in data[data_type]:
                if (
                    (data_item['validity']['from'] ==
                     item_payload['validity']['from']) and

                    (data_item['validity']['to'] ==
                     item_payload['validity']['to']) and

                    (data_item['value'] == item_payload['value'])
                ):
                    found_hit = True
                    logger.debug('Found hit in adress')
        elif data_type == 'manager':
            for data_item in data[data_type]:
                identical = self._std_compare(item_payload, data_item,
                                              'manager_level')
                uuids = []
                for item in item_payload['responsibility']:
                    uuids.append(item['uuid'])
                for responsibility in data_item['responsibility']:
                    identical = identical and (responsibility['uuid'] in uuids)
                identical = (identical and
                             (len(data_item['responsibility']) == len(uuids)))
                if identical:
                    found_hit = True
                    logger.debug('Found  hit in manager')
        elif data_type == 'association':
            for data_item in data[data_type]:
                if self._std_compare(item_payload, data_item, 'association_type'):
                    found_hit = True
                    logger.debug('Found hit in association')
        else:
            raise Exception('Uknown detail!')
        logger.info('Found hit: {}'.format(found_hit))
        return found_hit
